import logging
import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib.colors import ListedColormap
import numpy as np
from numpy.linalg import pinv
import pandas as pd
from scipy.interpolate import interp1d
from scipy.optimize import linprog
from scipy.sparse import coo_matrix

# from aggregate.utilities import FigureManager
from . import Portfolio, Aggregate, Distortion, Underwriter, FigureManager
from . constants import *

logger = logging.getLogger(__name__)


class Bounds(object):
    """
    Implement IME 2022 pricing bounds methodology.

    Typical usage: First, create a Portfolio or Aggregate object a. Then ::

        bd = cd.Bounds(a)
        bd.tvar_cloud('line', premium=, a=, n_tps=, s=, kind=)
        p_star = bd.p_star('line', premium)
        bd.cloud_view(axes, ...)

    :param distribution_spec: A Portfolio or Portfolio.density_df dataframe or pd.Series (must have loss as index)
            If DataFrame or Series values interpreted as desnsity, sum to 1. F, S, exgta all computed using Portfolio
            methdology
            If DataFrame line --> p_{line}
    """
    # from common_scripts.cs

    def __init__(self, distribution_spec):
        assert isinstance(distribution_spec, (pd.Series,
                          pd.DataFrame, Portfolio, Aggregate))
        self.distribution_spec = distribution_spec
        # although passed as input to certain functions (tvar with bounds) b is actually fixed
        self.b = 0
        self.Fb = 0
        self.tvar_function = None
        self.tvars = None
        self.tps = None
        self.weight_df = None
        self.idx = None
        self.hinges = None
        # in cases where we hold the tvar function here
        self._tail_var = None
        self._inverse_tail_var = None
        self.cloud_df = None
        # uniform mode
        self._t_mode = 'u'
        # data frame with tvar weights and principal extreme distortion weights by method
        self.pedw_df = None
        self._tvar_df = None
        # hack for beta distribution, you want to force 1 to be in tvar ps, but Fp = 1
        # TODO figure out why p_star grinds to a halt if you input b < inf
        self.add_one = False

    def __repr__(self):
        """
        Gets called automatically but so we can tweak.

        :return:
        """
        return 'My Bounds Object at ' + super(Bounds, self).__repr__()

    def __str__(self):
        return 'Hello' + super(Bounds, self).__repr__()

    @property
    def tvar_df(self):
        if self._tvar_df is None:
            self._tvar_df = pd.DataFrame(
                {'p': self.tps, 'tvar': self.tvars}).set_index('p')
        return self._tvar_df

    @property
    def t_mode(self):
        return self._t_mode

    @t_mode.setter
    def t_mode(self, val):
        assert val in ['u', 'gl']
        self._t_mode = val

    def make_tvar_function(self, line, b=np.inf):
        """
        make the tvar function from a Series p_total indexed by loss
        Includes determining sup and putting in value for zero
        If sup is largest value in index, sup set to inf

        also sets self.Fb

        Applies to min(Line, b)

        :param line:
        :param b:  bound on the losses, e.g., to model limited liability insurer
        :return:
        """
        self.b = b
        if isinstance(self.distribution_spec, Portfolio):
            assert line in self.distribution_spec.line_names_ex
            if line == 'total':
                self.tvar_function = self.distribution_spec.tvar
                self.Fb = self.distribution_spec.cdf(b)
            else:
                ag = getattr(self.distribution_spec, line)
                self.tvar_function = ag.tvar
                self.Fb = ag.cdf(b)
            if np.isinf(b):
                self.Fb = 1.0
            return

        elif isinstance(self.distribution_spec, Aggregate):
            self.tvar_function = self.distribution_spec.tvar
            self.Fb = self.distribution_spec.cdf(b)
            if np.isinf(b):
                self.Fb = 1.0
            return

        elif isinstance(self.distribution_spec, pd.DataFrame):
            assert f'p_{line}' in self.distribution_spec.columns
            # given a port.density_df
            p_total = self.distribution_spec[f'p_{line}']

        elif isinstance(self.distribution_spec, pd.Series):
            logger.info('tvar_array using Series')
            p_total = self.distribution_spec

        # need to create tvar function on the fly, using same method as Portfolio and Aggregate:
        bs = p_total.index[1]
        F = p_total.cumsum()
        if np.isinf(b):
            self.Fb = 0
        else:
            self.Fb = F[b]

        S = p_total.shift(-1, fill_value=min(p_total.iloc[-1], max(
            0, 1. - (p_total.sum()))))[::-1].cumsum()[::-1]
        lev = S.shift(1, fill_value=0).cumsum() * bs
        ex1 = lev.iloc[-1]
        ex = np.sum(p_total * p_total.index)
        logger.info(
            f'Computed mean loss for {line} = {ex:,.15f} (diff {ex - ex1:,.15f}) max F = {max(F)}')
        exgta = (ex - lev) / S + S.index
        sup = (p_total[::-1] > 0).idxmax()
        if sup == p_total.index[-1]:
            sup = np.inf
        exgta[S == 0] = sup
        logger.info(
            f'sup={sup}, max = {(p_total[::-1] > 0).idxmax()} "inf" = {p_total.index[-1]}')

        def _tvar(p, kind='interp'):
            """
            UNLIMITED tvar function!
            :param p:
            :param kind:
            :return:
            """
            if kind == 'interp':
                # original implementation interpolated
                if self._tail_var is None:
                    # make tvar function
                    self._tail_var = interp1d(F, exgta, kind='linear', bounds_error=False,
                                              fill_value=(0, sup))
                return self._tail_var(p)
            elif kind == 'inverse':
                if self._inverse_tail_var is None:
                    # make tvar function
                    self._inverse_tail_var = interp1d(exgta, F, kind='linear', bounds_error=False,
                                                      fill_value='extrapolate')
                return self._inverse_tail_var(p)

        self.tvar_function = _tvar

    def make_ps(self, n, mode):
        """
        Mode are you making s points (always uniform) or tvar p points (use t_mode)?
        self.t_mode == 'u': make uniform s points against which to evaluate g from 0 to 1 inclusive with more around 0
        self.t_mode == 'gl': make Gauss-Legndre p points at which TVaRs are evaluated from 0 inclusive to 1 exclusive with more around 1

        :param n:
        :return:
        """
        assert mode in ('s', 't')

        if mode == 't' and (self.Fb < 1 or self.add_one):
            # we will add 1 at the end
            n -= 1

        # Gauus Legendre points
        lg = np.polynomial.legendre.leggauss

        if self.t_mode == 'gl':
            if mode == 's':
                x, wts = lg(n - 2)
                ps = np.hstack((0, (x + 1) / 2, 1))

            elif mode == 't':
                x, wts = lg(n * 2 + 1)
                ps = x[n:]
        elif self.t_mode == 'u':
            if mode == 's':
                ps = np.linspace(1 / n, 1, n)
            elif mode == 't':
                # exclude 1 (sup distortion) at the end; 0=mean
                ps = np.linspace(0, 1, n, endpoint=False)
        # always ensure that 1  is in ps for t mode when b < inf if Fb < 1
        if mode == 't' and self.Fb < 1 or self.add_one:
            ps = np.hstack((ps, 1))
        return ps

    def tvar_array(self, line, n_tps=256, b=np.inf, kind='interp'):
        """
        Compute tvars at n equally spaced points, tps.


        :param line:
        :param n_tps:  number of tvar p points, default 256
        :param b: cap on losses applied before computing TVaRs (e.g., adjust losses for finite assets b).
               Use np.inf for unlimited losses.
        :param kind: if interp  uses the standard function, easy, for continuous distributions; if 'tail' uses
               explicit integration of tail values, for discrete distributions
        :return:
        """
        assert kind in ('interp', 'tail')
        self.make_tvar_function(line, b)

        logger.info(f'F(b) = {self.Fb:.5f}')
        # tvar p values should linclude 0 (the mean) but EXCLUDE 1
        # self.tps = np.linspace(0.5 / n_tps, 1 - 0.5 / n_tps, n_tps)
        self.tps = self.make_ps(n_tps, 't')

        if kind == 'interp':
            self.tvars = self.tvar_function(self.tps)
            if not np.isinf(b):
                # subtract S(a)(TVaR(F(a)) - a)
                # do all at once here - do not call self.tvar_with_bounds function
                self.tvars = np.where(self.tps <= self.Fb,
                                      self.tvars -
                                      (1 - self.Fb) * (self.tvar_function(self.Fb) -
                                                       b) / (1 - self.tps),
                                      b)
        elif kind == 'tail':
            self.tvars = np.array(
                [self.tvar_with_bound(i, b, kind) for i in self.tps])

    def p_star(self, line, premium, b=np.inf, kind='interp'):
        """
        Compute p* so TVaR @ p* of min(X, b) = premium

        In this case the cap b has an impact (think of integrating q(p) over p to 1, q is impacted by b)

        premium <= b is required (no rip off condition)

        If b < inf then must solve TVaR(p) - (1 - F(b)) / (1 - p)[TVaR(F(b)) - b] = premium
        Let k = (1 - F(b)) [TVaR(F(b)) - b], so solving

        f(p) = TVaR(p) - k / (1 - p) - premium == 0

        using NR

        :param line:
        :param premium: target premium
        :param b:  bound
        :return:
        """
        assert kind in ('interp', 'tail')
        if premium > b:
            raise ValueError(
                f'p_star must have premium ({premium}) <= largest loss bound ({b})')

        if kind == 'interp':

            self.make_tvar_function(line, b)

            if np.isinf(b):
                p_star = self.tvar_function(premium, 'inverse')
            else:
                # nr, remember F(a) is self.Fa set by make_tvar_function
                k = (1 - self.Fb) * (self.tvar_function(self.Fb) - b)

                def f(p):
                    return self.tvar_function(p) - k / (1 - p) - premium

                # should really compute f' numerically, but...
                fp = 100
                p = 0.5
                iters = 0
                delta = 1e-10
                while abs(fp) > 1e-6 and iters < 20:
                    fp = f(p)
                    fpp = (f(p + delta) - f(p)) / delta
                    pnew = p - fp / fpp
                    if 0 <= pnew <= 1:
                        p = pnew
                    elif pnew < 0:
                        p = p / 2
                    else:
                        #  pnew > 1:
                        p = (1 + p) / 2

                if iters == 20:
                    logger.warning(
                        f'Questionable convergence solving for p_star, last error {fp}.')
                p_star = p

        elif kind == 'tail':
            def f(p):
                return self.tvar_with_bound(p, b, 'tail') - premium

            fp = 100
            p = 0.5
            iters = 0
            delta = 1e-10
            while abs(fp) > 1e-6 and iters < 20:
                fp = f(p)
                fpp = (f(p + delta) - f(p)) / delta
                pnew = p - fp / fpp
                if 0 <= pnew <= 1:
                    p = pnew
                elif pnew < 0:
                    p = p / 2
                else:
                    #  pnew > 1:
                    p = (1 + p) / 2

            if iters == 20:
                logger.warning(
                    f'Questionable convergence solving for p_star, last error {fp}.')
            p_star = p

        return p_star

    def tvar_with_bound(self, p, b=np.inf, kind='interp'):
        """
        Compute tvar taking bound into account.
        Assumes tvar_function setup.

        Warning: b must equal the b used when calibrated. The issue is computing F
        varies with the type of underlying portfolio. This is fragile.
        Added storing b and checking equal. For backwards comp. need to keep b argument

        :param p:
        :param b:
        :return:
        """
        assert self.tvar_function is not None
        assert b == self.b

        if kind == 'interp':
            tvar = self.tvar_function(p)
            if not np.isinf(b):
                if p < self.Fb:
                    tvar = tvar - (1 - self.Fb) * \
                        (self.tvar_function(self.Fb) - b) / (1 - p)
                else:
                    tvar = b
        elif kind == 'tail':
            # use the tail method for discrete distributions
            tvar = self.distribution_spec.tvar(p, 'tail')
            if not np.isinf(b):
                if p < self.Fb:
                    tvar = tvar - \
                        (1 - self.Fb) * \
                        (self.distribution_spec.tvar(self.Fb, 'tail') - b) / (1 - p)
                else:
                    tvar = b
        return tvar

    def compute_weight(self, premium, p0, p1, b=np.inf, kind='interp'):
        """
        compute the weight for a single TVaR p0 < p1 value pair

        :param line:
        :param premium:
        :param tp:
        :param b:
        :return:
        """

        assert p0 < p1
        assert self.tvar_function is not None

        lhs = self.tvar_with_bound(p0, b, kind)
        rhs = self.tvar_with_bound(p1, b, kind)

        assert lhs != rhs
        weight = (premium - lhs) / (rhs - lhs)
        return weight

    def compute_weights(self, line, premium, n_tps, b=np.inf, kind='interp'):
        """
        Compute the weights of the extreme distortions

        Applied to min(line, b)  (allows to work for net)

        Note: independent of the asset level

        :param line: within port, or total
        :param premium: target premium for the line
        :param n_tps: number of tvar p points (tps)number of tvar p points (tps)number of tvar p points
            (tps)number of tvar p points (tps).
        :param b: loss bound: compute weights for min(line, b); generally used for net losses only.
        :return:
        """

        self.tvar_array(line, n_tps, b, kind)
        # you add zero, so there will be one additional point
        # n_tps += 1
        p_star = self.p_star(line, premium, b, kind)
        if p_star in self.tps:
            logger.critical('p_star in tps')
            # raise ValueError()

        lhs = self.tps[self.tps <= p_star]
        rhs = self.tps[self.tps > p_star]

        tlhs = self.tvars[self.tps <= p_star]
        trhs = self.tvars[self.tps > p_star]

        lhs, rhs = np.meshgrid(lhs, rhs)
        tlhs, trhs = np.meshgrid(tlhs, trhs)

        df = pd.DataFrame({'p_lower': lhs.flat, 'p_upper': rhs.flat,
                           't_lower': tlhs.flat, 't_upper': trhs.flat,
                           })
        # will fail when p_star in self.ps; let's deal with then when it happens
        df['weight'] = (premium - df.t_lower) / (df.t_upper - df.t_lower)

        df = df.set_index(['p_lower', 'p_upper'], verify_integrity=True)
        df = df.sort_index()

        if p_star in self.tps:
            # raise ValueError('Found pstar in ps')
            logger.critical(f'Found p_star = {p_star} in ps!!')
            df.at[(p_star, p_star), 'weight'] = 1.0

        logger.info(f'p_star={p_star:.4f}, len(p<=p*) = {len(df.index.levels[0])}, '
                    f'len(p>p*) = {len(df.index.levels[1])}; '
                    f' pstar in ps: {p_star in self.tps}')

        self.weight_df = df

        # index for tp values
        r = np.arange(n_tps)
        r_rhs, r_lhs = np.meshgrid(r[self.tps > p_star], r[self.tps <= p_star])
        self.idx = np.vstack((r_lhs.flat, r_rhs.flat)).reshape((2, r_rhs.size))

    def tvar_hinges(self, s):
        """
        make the tvar hinge functions by evaluating each tvar_p(s) = min(1, s/(1-p) for p in tps, at EP points s

        all arguments in [0,1] x [0,1]

        :param s:
        :return:
        """

        self.hinges = coo_matrix(np.minimum(1.0, s.reshape(
            1, len(s)) / (1.0 - self.tps.reshape(len(self.tps), 1))))

    def tvar_cloud(self, line, premium, a, n_tps, s, kind='interp'):
        """
        weight down tvar functions to the extremal convex measures

        asset level a acts like an agg stop on what is being priced, i.e. we are working with min(X, a)

        :param line:
        :param premium:
        :param a:
        :param n_tps:
        :param s:
        :param b:  bound, applies to min(line, b)
        :return:
        """

        self.compute_weights(line, premium, n_tps, a, kind)

        if type(s) == int:
            # points at which g is evaluated - all OK to include 0 and 1
            # s = np.linspace(0, 1, s+1, endpoint=True)
            s = self.make_ps(s, 's')

        self.tvar_hinges(s)

        ml = coo_matrix((1 - self.weight_df.weight, (np.arange(len(self.weight_df)), self.idx[0])),
                        shape=(len(self.weight_df), len(self.tps)))
        mr = coo_matrix((self.weight_df.weight, (np.arange(len(self.weight_df)), self.idx[1])),
                        shape=(len(self.weight_df), len(self.tps)))
        m = ml + mr

        logger.info(
            f'm shape = {m.shape}, hinges shape = {self.hinges.shape}, types {type(m)}, {type(self.hinges)}')

        self.cloud_df = pd.DataFrame(
            (m @ self.hinges).T.toarray(), index=s, columns=self.weight_df.index)
        self.cloud_df.index.name = 's'

    def cloud_view(self, axs, n_resamples, scale='linear', alpha=0.05, pricing=True, distortions=None,
                   title='', lim=(-0.025, 1.025), check=False, add_average=True):
        """
        visualize the cloud with n_resamples

        after you have recomputed...

        if there are distortions plot on second axis

        :param axs:
        :param n_resamples: if random sample
        :param scale: linear or return
        :param alpha: opacity
        :param pricing: restrict to p_max = 0, ensuring g(s)<1 when s<1
        :param distortions:
        :param title: optional title (applied to all plots)
        :param lim: axis limits
        :param check:   construct and plot Distortions to check working ; reduces n_resamples to 5
        :return:
        """
        assert scale in ['linear', 'return']
        assert not distortions or (len(axs.flat) > 1)
        bit = None
        if check:
            n_resamples = min(n_resamples, 5)
        norm = mpl.colors.Normalize(0, 1)
        cm = mpl.cm.ScalarMappable(norm=norm, cmap='viridis_r')
        mapper = cm.get_cmap()
        s = np.linspace(0, 1, 1001)

        def plot_max_min(ax):
            ax.fill_between(self.cloud_df.index, self.cloud_df.min(
                1), self.cloud_df.max(1), facecolor='C7', alpha=.15)
            self.cloud_df.min(1).plot(
                ax=ax, label='_nolegend_', lw=1, ls='-', c='k')
            self.cloud_df.max(1).plot(
                ax=ax, label="_nolegend_", lw=1, ls='-', c='k')

        logger.info('starting cloudview...')
        if scale == 'linear':
            ax = axs[0]
            if n_resamples > 0:
                if pricing:
                    bit = self.weight_df.xs(0, drop_level=False).sample(
                        n=n_resamples, replace=True).reset_index()
                else:
                    bit = self.weight_df.sample(
                        n=n_resamples, replace=True).reset_index()
                logger.info('cloudview...done 1')
                # display(bit)
                for i in bit.index:
                    pl, pu, tl, tu, w = bit.loc[i]
                    self.cloud_df[(pl, pu)].plot(
                        ax=ax, lw=1, c=mapper(w), alpha=alpha, label=None)
                    if check:
                        # put in actual for each sample
                        d = Distortion('wtdtvar', w, df=[pl, pu])
                        gs = d.g(s)
                        ax.plot(s, gs, c=mapper(w), lw=2, ls='--',
                                alpha=.5, label=f'ma ({pl:.3f}, {pu:.3f}) ')
                ax.get_figure().colorbar(cm, ax=ax, shrink=.5, aspect=16,
                                         label='Weight to Higher Threshold')
            else:
                logger.info('cloudview: no resamples, skipping 1')
            logger.info('cloudview: start max/min')
            plot_max_min(ax)
            logger.info('cloudview: done with max/min')
            for ln in ax.lines:
                ln.set(label=None)
            if check:
                ax.legend(loc='lower right', fontsize='large')
            ax.plot([0, 1], [0, 1], c='k', lw=.25, ls='-')
            ax.set(xlim=lim, ylim=lim, aspect='equal')

            if type(distortions) == dict:
                distortions = [distortions]
            if distortions == 'space':
                ax = axs[1]
                plot_max_min(ax)
                ax.plot([0, 1], [0, 1], c='k', lw=.25,
                        ls='-', label='_nolegend_')
                ax.legend(loc='lower right', ncol=3, fontsize='large')
                ax.set(xlim=lim, ylim=lim, aspect='equal')
            elif type(distortions) == list:
                logger.info('cloudview: start 4 adding distortions')
                name_mapper = {
                    'roe': 'CCoC', 'tvar': 'TVaR(p*)', 'ph': 'PH', 'wang': 'Wang', 'dual': 'Dual'}
                lss = list(mpl.lines.lineStyles.keys())
                for ax, dist_dict in zip(axs[1:], distortions):
                    ii = 1
                    for k, d in dist_dict.items():
                        gs = d.g(s)
                        k = name_mapper.get(k, k)
                        ax.plot(s, gs, lw=1, ls=lss[ii], label=k)
                        ii += 1
                    plot_max_min(ax)
                    ax.plot([0, 1], [0, 1], c='k', lw=.25,
                            ls='-', label='_nolegend_')
                    ax.legend(loc='lower right', ncol=3, fontsize='large')
                    ax.set(xlim=lim, ylim=lim, aspect='equal')
                if add_average:
                    self.cloud_df.mean(1).plot(ax=ax, c=f'C{len(distortions[-1])}',
                                               ls='-.', lw=.5, label='Avg extreme')
            else:
                # do nothing
                pass

        elif scale == 'return':
            ax = axs[0]
            bit = self.cloud_df.sample(n=n_resamples, axis=1)
            bit.index = 1 / bit.index
            bit = 1 / bit
            bit.plot(ax=ax, lw=.5, c='C7', alpha=alpha)
            ax.plot([0, 1000], [0, 1000], c='C0', lw=1)
            ax.legend().set(visible=False)
            ax.set(xscale='log', yscale='log')
            ax.set(xlim=[2000, 1], ylim=[2000, 1])

        if title != '':
            for ax in axs:
                if bit is not None:
                    title1 = f'{title}, n={len(bit)} samples'
                else:
                    title1 = title
                ax.set(title=title1)
        for ax in axs[1:]:
            ax.legend(ncol=1, loc='lower right')
        for ax in axs:
            ax.set(title=None)

    def weight_image(self, ax, levels=20, colorbar=True):
        bit = self.weight_df.weight.unstack()
        img = ax.contourf(bit.columns, bit.index, bit,
                          cmap='viridis_r', levels=levels)
        ax.set(xlabel='p1', ylabel='p0', title='Weight for p1', aspect='equal')
        if colorbar:
            ax.get_figure().colorbar(img, ax=ax, shrink=.5, aspect=16, label='Weight to p_upper')

    def quick_price(self, distortion, a):
        """
        price total to assets a using distortion

        requires distribution_spec has a density_df dataframe with a p_total or p_total

        TODO: add ability to price other lines
        :param distortion:
        :param a:
        :return:
        """

        if isinstance(self.distribution_spec, (Portfolio, Aggregate)):
            df = self.distribution_spec.density_df
            bs = self.distribution_spec.bs
        elif isinstance(self.distribution_spec, pd.DataFrame):
            df = self.distribution_spec
            bs = df.index[1]
        else:
            raise NotImplemented('Must input Aggregate, Portfolio, or DataFrame, '
                                 f'not type {type(self.distribution_spec)}')

        temp = distortion.g(
            df.p_total.shift(-1, fill_value=0)[::-1].cumsum())[::-1]

        if isinstance(temp, np.ndarray):
            # not all g functions return Series (you can't guarantee it is called on something with an index)
            temp = pd.Series(temp, index=df.index)

        temp = temp.shift(1, fill_value=0).cumsum() * bs

        if np.isinf(a):
            r = len(temp) - 1
        else:
            r = temp.index.get_loc(a)

        return temp.iloc[r] * bs

    def principal_extreme_distortion_analysis(self, gs, pricing=False):
        """
        Find the principal extreme distortion analysis to solve for gs = g(s), s=self.cloud_df.index

        Assumes that tvar_cloud has been called and that cloud_df exists
        len(gs) = len(cloud_df)

        E.g., call

            b = Bounds(port)
            b.t_mode = 'u'
            # set premium and asset level a
            b.tvar_cloud('total', premium, a)
            # make gs
            b.principal_extreme_distortion_analysis(gs)

        :param gs: either g(s) evaluated on s = cloud_df.index or the name of a calibrated distortion in
            distribution_spec.dists (created by a call to calibrate_distortions)
        :param pricing: if try, try just using pricing distortions
        :return:
        """

        assert self.cloud_df is not None

        if type(gs) == str:
            s = np.array(self.cloud_df.index)
            gs = self.distribution_spec.dists[gs].g(s)

        assert len(gs) == len(self.cloud_df)

        if pricing:
            _ = self.cloud_df.xs(0, axis=1, level=0, drop_level=False)
            X = _.to_numpy()
            idx = _.columns
        else:
            _ = self.cloud_df
            X = _.to_numpy()
            idx = _.columns
        n = X.shape[1]

        print(X.shape, self.cloud_df.shape)

        # Moore Penrose solution
        mp = pinv(X) @ gs
        logger.info('Moore-Penrose solved...')

        # optimization solutions
        A = np.hstack((X, np.eye(X.shape[0])))
        b_eq = gs
        c = np.hstack((np.zeros(X.shape[1]), np.ones_like(b_eq)))

        lprs = linprog(c, A_eq=A, b_eq=b_eq, method='revised simplex')
        logger.info(
            f'Revised simpled solved...\nSum of added variables={np.sum(lprs.x[n:])} (should be zero for exact)')
        self.lprs = lprs

        lpip = linprog(c, A_eq=A, b_eq=b_eq, method='interior-point')
        logger.info(
            f'Interior point solved...\nSum of added variables={np.sum(lpip.x[n:])}')
        self.lpip = lpip

        print(lprs.x, lpip.x)

        # consolidate answers
        self.pedw_df = pd.DataFrame(
            {'w_mp': mp, 'w_rs': lprs.x[:n], 'w_ip': lpip.x[:n]}, index=idx)
        self.pedw_df['w_upper'] = self.weight_df.weight

        # diagnostics
        for c in self.pedw_df.columns[:-1]:
            answer = self.pedw_df[c].values
            ganswer = answer[answer > 1e-16]
            logger.info(
                f'Method {c}\tMinimum parameter {np.min(answer)}\tNumber non-zero {len(ganswer)}')

        return gs

    def ped_distortion(self, n, solver='rs'):
        """
        make the approximating distortion from the first n Principal Extreme Distortions (PED)s using rs or ip solutions

        :param n:
        :return:
        """
        assert solver in ['rs', 'ip']

        # the weight column for solver
        c = f'w_{solver}'
        # pull off the tvar and PED weights
        df = self.pedw_df.sort_values(c, ascending=False)
        bit = df.loc[df.index[:n], [c, 'w_upper']]
        # re-weight partial (method / lp-solve) weights to 1
        bit[c] /= bit[c].sum()
        # multiply lp-solve weights with the weigh_df extreme distortion p_lower/p_upper weights
        bit['c_lower'] = (1 - bit.w_upper) * bit[c]
        bit['c_upper'] = bit.w_upper * bit[c]
        # gather into data frame of p and total weight (labeled c)
        bit2 = bit.reset_index().drop([c, 'w_upper'], 1)
        bit2.columns = bit2.columns.str.split('_', expand=True)
        bit2 = bit2.stack(1).groupby('p')['c'].sum()
        # bit2 has index = probability points and values = weights for the wtd tvar distortion
        d = Distortion.wtd_tvar(bit2.index, bit2.values, f'PED({solver}, {n})')
        return d


def similar_risks_graphs_sa(axd, bounds, port, pnew, roe, prem, p_reg=1):
    """
    stand-alone
    ONLY WORKS FOR BOUNDED PORTFOLIOS (use for beta mixture examples)
    Updated version in CaseStudy
    axd from mosaic
    bounds = Bounds class from port (calibrated to some base)it
    pnew = new portfolio
    input new beta(a,b) portfolio, using existing bounds object

    sample: see similar_risks_sample()

    Provenance : from make_port in Examples_2022_post_publish
    """

    if axd is None:
        fig = plt.figure(constrained_layout=True, figsize=(12, 6))
        axd = fig.subplot_mosaic(
            '''
            AAAABBFF
            AAAACCFF
            AAAADDEE
            AAAADDEE
        ''')

    df = bounds.weight_df.copy()
    df['test'] = df['t_upper'] * df.weight + df.t_lower * (1 - df.weight)

    # HERE IS ISSUE - should really use tvar with bounds and incorporate the bound
    if p_reg < 1:
        logger.warning('figuring tvars with bounds')
        btemp = Bounds(pnew)
        b = pnew.q(p_reg)
        btemp.make_tvar_function('total', b=b)
        tvar1 = {p: btemp.tvar_with_bound(p, b=b) for p in bounds.tps}
    else:
        tvar1 = {p: float(pnew.tvar(p)) for p in bounds.tps}
    df['t1_lower'] = [tvar1[p] for p in df.index.get_level_values(0)]
    df['t1_upper'] = [tvar1[p] for p in df.index.get_level_values(1)]
    df['t1'] = df.t1_upper * df.weight + df.t1_lower * (1 - df.weight)

    roe_d = Distortion('roe', roe)
    tvar_d = Distortion('tvar', bounds.p_star('total', prem))
    idx = df.index.get_locs(df.idxmax()['t1'])[0]
    pl, pu, tl, tu, w = df.reset_index().iloc[idx, :-4]
    max_d = Distortion('wtdtvar', w, df=[pl, pu])

    tmax = float(df.iloc[idx]['t1'])
    n_ = len(df.query('t1 == @tmax'))
    logger.warning(f'Ties for max: {n_}')
    n_ = len(df.query(f't1 >= {tmax} - 1e-4'))
    logger.warning(f'Near ties for max: {n_}')

    idn = df.index.get_locs(df.idxmin()['t1'])[0]
    pln, pun, tl, tu, wn = df.reset_index().iloc[idn, :-4]
    min_d = Distortion('wtdtvar', wn, df=[pln, pun])

    ax = axd['A']
    plot_max_min(bounds, ax)
    n = len(ax.lines)
    roe_d.plot(ax=ax, both=False)
    tvar_d.plot(ax=ax, both=False)
    max_d.plot(ax=ax, both=False)
    min_d.plot(ax=ax, both=False)

    ax.lines[n + 0].set(label='roe', color='C0', ls='--')
    ax.lines[n + 2].set(color='C1', label='tvar', ls='-.')
    ax.lines[n + 4].set(color='C4', label='max', lw=1)
    ax.lines[n + 6].set(color='C5', label='min', lw=1)
    # the average
    bounds.cloud_df.mean(1).plot(ax=ax, c='C3',
                               ls='-.', lw=1.5, label='Avg extreme')
    ax.legend(loc='upper left')

    ax.set(title=f'Max ({pl}, {pu}), min ({pln}, {pun})')

    ax = axd['B']
    bounds.weight_image(ax)

    bit = df['t1'].unstack(1)
    ax = axd['C']
    img = ax.contourf(bit.columns, bit.index, bit, cmap='viridis_r', levels=20)
    ax.set(xlabel='p1', ylabel='p0', title='Pricing on New Risk', aspect='equal')
    ax.get_figure().colorbar(img, ax=ax, shrink=.5, aspect=16, label='rho(X_new)')
    ax.plot(pu, pl, '.', c='w')
    ax.plot(pun, pln, 's', ms=3, c='white')

    ax = axd['D']
    plot_lee(port, ax, 'C0', lw=1)
    plot_lee(pnew, ax, 'C1')
    ax.set(ylim=[0, port.q(0.999)])
    ax.legend()

    ax = axd['E']
    port.density_df.p_total.plot(ax=ax, logy=True, lw=1, label=port.name)
    pnew.density_df.p_total.plot(ax=ax, logy=True, lw=1, label=pnew.name)
    ax.legend()
    ax.set(title='Total, log densities')

    ax = axd['F']
    plot_max_min(bounds, ax)
    for c, dd in zip(['C0', 'C1', 'C2'], ['ph', 'wang', 'dual']):
        port.dists[dd].plot(ax=ax, both=False, lw=1)
        ax.lines[n].set(c=c, label=dd)
        n += 2
    ax.legend(loc='lower right')

    return df


def similar_risks_example():
    """
    Interesting beta risks and how to use similar_risks_sa.


    :return:
    """
    # stand alone hlep from the code; split at program = to run different options
    uw = Underwriter()
    p_base = uw.write('''
    port UNIF
        agg ONE 1 claim sev 1 * beta 1 1 fixed
    ''')
    p_base.update(11, 1 / 1024, remove_fuzz=True)
    prem = p_base.tvar(0.2, 'interp')
    a = 1
    d = (prem - p_base.ex) / (a - p_base.ex)
    v = 1 - d
    roe = d / v
    prem, roe
    p_base.calibrate_distortions(As=[1], ROEs=[roe], strict='ordered')
    bounds = Bounds(p_base)
    bounds.tvar_cloud('total', prem, a, 128 * 2, 64 * 2, 'interp')
    p_star = bounds.p_star('total', prem, kind='interp')

    smfig = FigureManager(cycle='c', color_mode='color', font_size=10, legend_font='small',
                          default_figsize=(FIG_W, FIG_H))

    f, axs = smfig(1, 3, (18.0, 6.0), )
    ax0, ax1, ax2 = axs.flat
    axi = iter(axs.flat)
    # all with base portfolio

    bounds.cloud_view(axs.flatten(), 0, alpha=1, pricing=True,
                      title=f'Premium={prem:,.1f}, a={a:,.0f}, p*={p_star:.3f}',
                      distortions=[{k: p_base.dists[k] for k in ['roe', 'tvar']},
                                   {k: p_base.dists[k] for k in ['ph', 'wang', 'dual']}])
    for ax in axs.flatten()[1:]:
        ax.legend(ncol=1, loc='lower right')
    for ax in axs.flatten():
        ax.set(title=None)

    program = '''
    port BETA
        agg TWO 1 claim sev 1 * beta [200 300 400 500 600 7] [600 500 400 300 200 1] wts=6 fixed
        # never worked
        # agg TWO 1 claim sev 1 * beta [1 2000 4000 6000 50] [100 6000 4000 2000 1] wts[0.1875 0.1875 0.1875 0.1875 .25] fixed
        # interior solution:
        # agg TWO 1 claim sev 1 * beta [300 400 500 600 35] [500 400 300 200 5] wts[.125 .25 .125 .25 .25] fixed
        #
        # agg TWO 1 claim sev 1 * beta [50 30 1] [1 40 10] wts=3 fixed
        # agg TWO 1 claim sev 1 * beta [50 30 1] [1 40 10] wts[.375 .375 .25] fixed

    '''
    p_new = uw.write(program)
    p_new.update(11, 1 / 1024, remove_fuzz=True)

    p_new.plot(figsize=(6, 4))

    axd = plt.figure(constrained_layout=True, figsize=(16, 8)).subplot_mosaic(
        '''
        AAAABBFF
        AAAACCFF
        AAAADDEE
        AAAADDEE
    '''
    )
    df = similar_risks_graphs_sa(axd, bounds, p_base, p_new, roe, prem)
    return df


def plot_max_min(self, ax):
    """
    Extracted from bounds, self=Bounds object
    """
    ax.fill_between(self.cloud_df.index, self.cloud_df.min(
        1), self.cloud_df.max(1), facecolor='C7', alpha=.15)
    self.cloud_df.min(1).plot(ax=ax, label='_nolegend_', lw=0.5, ls='-', c='k')
    self.cloud_df.max(1).plot(ax=ax, label="_nolegend_", lw=0.5, ls='-', c='k')


def plot_lee(port, ax, c, lw=1):
    """
    Lee diagram by hand
    """
    p_ = np.linspace(0, 1, 1001)
    qs = [port.q(p) for p in p_]
    ax.step(p_, qs, lw=lw, c=c, label=port.name)
    ax.set(xlim=[-0.05, 1.05], ylim=[-0.05, max(qs) + .05],
           title=f'{port.name} Lee diagram')



# make the cat non-cat case study

from aggregate.extensions import case_studies as cs
from aggregate import build
import warnings

warnings.filterwarnings('ignore')

if __name__ == '__main__':
    # be more informative
    build.logger_level(20)
    # calibrate reinsurance
    recalc = build('agg Cat 1 claim sev lognorm 20 cv 1.00 fixed'
                   , log2=16, bs=1/64)
    a, d = recalc.q(0.9), recalc.q(0.995)
    y = d - a

    # create exhibits
    cnc = cs.CaseStudy()
    cnc.factory(case_id='cnc',
                case_name='Cat/Non-Cat',
                case_description='Cat/Non-Cat in the new syntax.',
                a_distribution='agg NonCat 1 claim sev gamma 80 cv 0.15 fixed',
                b_distribution_gross='agg Cat 1 claim sev lognorm 20 cv 1.00 fixed',
                b_distribution_net=f'agg Cat 1 claim sev lognorm 20 cv 1.00 fixed aggregate net of {y} xs {a}',
                reg_p=0.999,
                roe=0.10,
                d2tc=0.3,
                f_discrete=False,
                s_values=[.005, 0.01, 0.03],
                gs_values=[0.029126,   0.047619,   0.074074],
                bs=1/64,
                log2=16,
                padding=1)
    cnc.full_monty()
    cnc.to_json()
    cnc.browse_exhibits()



import numpy as np
import matplotlib.pyplot as plt
import scipy.stats as ss
from scipy.interpolate import interp1d
from scipy.spatial import ConvexHull
from io import StringIO
import pandas as pd
from textwrap import fill
import logging

from .constants import *

logger = logging.getLogger(__name__)


class Distortion(object):
    """
    Creation and management of distortion functions.

    0.9.4: renamed roe to ccoc, but kept creator with roe for backwards compatibility.
    Oct 2022: renamed wtdtvar to bitvar, but kept ...

    """

    # make these (mostly) immutable...avoid changing by mistake
    _available_distortions_ = ('ph', 'wang', 'cll', 'lep', 'ly', 'clin', 'dual', 'ccoc', 'tvar',
                               'bitvar', 'convex', 'tt')
    _has_mass_ = ('ly', 'clin', 'lep', 'roe')
    _med_names_ = ("Prop Hzrd", "Wang", 'Capd Loglin', "Lev Equiv", "Lin Yield", "Capped Linear", "Dual Mom",
                   'Const CoC', "Tail VaR", 'BiTVaR', "Convex Env", 'Wang-tt')
    _long_names_ = ("Proportional Hazard", "Wang-normal", 'Capped Loglinear', "Leverage Equivalent Pricing",
                    "Linear Yield", "Capped Linear", "Dual Moment", "Constant CoC", "Tail VaR",
                    'BiTVaR', "Convex Envelope", 'Wang-tt')
    # TODO fix examples!
    # _available_distortions_ = ('ph', 'wang', 'cll', 'lep',  'ly', 'clin', 'dual', 'ccoc', 'tvar', 'wtdtvar,  'convex')
    _eg_param_1_ =              (.9,     .1,      .9,    0.25,  0.8,   1.1,   1.5,   .1,     0.15,     .15)
    _eg_param_2_ =              (.5,     .75,     .5,    0.5,   1.5,   1.8,     3,   .25,    0.50,      .5)
    # _distortion_names_ = dict(zip(_available_distortions_, _med_names_))
    _distortion_names_ = dict(zip(_available_distortions_, _long_names_))
    renamer = _distortion_names_


    @classmethod
    def available_distortions(cls, pricing=True, strict=True):
        """
        List of the available distortions.

        :param pricing: only return list suitable for pricing, excludes tvar and convex
        :param strict: only include those without mass at zero  (pricing only)
        :return:
        """

        if pricing and strict:
            return tuple((i for i in cls._available_distortions_[:-4] if i not in cls._has_mass_))
        elif pricing:
            return cls._available_distortions_[:-2]
        else:
            return cls._available_distortions_

    def __init__(self, name, shape, r0=0.0, df=None, col_x='', col_y='', display_name=''):
        """
        Create a new distortion.

        Tester: ::

            ps = np.linspace(0, 1, 201)
            for dn in agg.Distortion.available_distortions(True):
                if dn=='clin':
                    # shape param must be > 1
                    g_dist = agg.Distortion(**{'name': dn, 'shape': 1.25, 'r0': 0.02, 'df': 5.5})
                else:
                    g_dist = agg.Distortion(**{'name': dn, 'shape': 0.5, 'r0': 0.02, 'df': 5.5})
                g_dist.plot()
                g = g_dist.g
                g_inv = g_dist.g_inv

                df = pd.DataFrame({'p': ps, 'gg_inv': g(g_inv(ps)), 'g_invg': g_inv(g(ps)),
                'g': g(ps), 'g_inv': g_inv(ps)})
                print(dn)
                print("errors")
                display(df.query(' abs(gg_inv - g_invg) > 1e-5'))

        :param name: name of an available distortion, call ``Distortion.available_distortions()`` for a list
        :param shape: float or [float, float]
        :param shape: shape parameter
        :param r0: risk free or rental rate of interest
        :param df:  for convex envelope, dataframe with col_x and col_y used to parameterize or df for t
        :param col_x:
        :param col_y:
        :param display_name: over-ride name, useful for parameterized convex fix distributions
        """

        self._name = name
        self.shape = shape
        self.r0 = r0
        # when created by calibrate distortions extra info put here
        self.error = 0.0
        self.premium_target = 0.0
        self.assets = 0.0
        self.mass = 0.0
        self.df = df
        self.col_x = col_x
        self.col_y = col_y
        self.display_name = display_name
        g_prime = None

        # now make g and g_inv
        if self._name == 'ph':
            rho = self.shape
            rhoinv = 1.0 / rho
            self.has_mass = False

            # @numba.vectorize(["float64(float64)"], nopython=True, target='parallel')
            def g(x):
                return x ** rho

            def g_inv(x):
                return x ** rhoinv

            def g_prime(x):
                return rho * x ** (rho - 1.0) if x > 0 else np.inf

        elif self._name == 'wang':
            lam = self.shape
            n = ss.norm()
            self.has_mass = False

            def g(x):
                return n.cdf(n.ppf(x) + lam)

            def g_inv(x):
                return n.cdf(n.ppf(x) - lam)

            def g_prime(x):
                return n.pdf(n.ppf(x) - lam) / n.pdf(n.ppf(x) + lam)

        elif self._name == 'tt':
            lam = self.shape
            t = ss.t(self.df)
            self.has_mass = False

            def g(x):
                return t.cdf(t.ppf(x) + lam)

            def g_inv(x):
                return t.cdf(t.ppf(x) - lam)

        elif self._name == 'cll':
            # capped log linear
            b = self.shape
            binv = 1 / b
            ea = np.exp(self.r0)
            a = self.r0
            self.has_mass = False

            def g(x):
                return np.where(x==0, 0, np.minimum(1, ea * x ** b))

            def g_inv(x):
                return np.where(x < 1, np.minimum(1, (x / ea) ** binv), 1)

        elif self._name == 'tvar':
            p = self.shape
            alpha = 1 / (1 - p)
            self.has_mass = False

            def g(x):
                return np.minimum(alpha * x, 1)

            def g_inv(x):
                return np.where(x < 1, x * (1 - p), 1)

            def g_prime(x):
                return np.where(x < 1 - p, alpha, 0)

        elif self._name == 'ly':
            # linear yield
            # r0 = occupancy; rk = consumption specified in list shape parameter
            rk = self.shape
            self.has_mass = (r0 > 0)
            self.mass = r0 / (1 + r0)

            def g(x):
                return np.where(x == 0, 0, (self.r0 + x * (1 + rk)) / (1 + self.r0 + rk * x))

            def g_inv(x):
                return np.maximum(0, (x * (1 + self.r0) - self.r0) / (1 + rk * (1 - x)))

        elif self._name == 'clin':
            # capped linear, needs shape > 1 to make sense...needs shape >= 1-r0 else
            # problems at 1
            sl = self.shape
            self.has_mass = (r0 > 0)
            self.mass = r0

            def g(x):
                return np.where(x == 0, 0, np.minimum(1, self.r0 + sl * x))

            def g_inv(x):
                return np.where(x <= self.r0, 0, (x - self.r0) / sl)

        elif self._name == 'roe' or self._name == 'ccoc':
            # constant roe = capped linear with shape = 1/(1+r), r0=r/(1+r)
            # r = target roe
            self._name = 'ccoc'
            r = self.shape
            v = 1 / (1 + r)
            d = 1 - v
            self.has_mass = (d > 0)
            self.mass = d

            def g(x):
                return np.where(x == 0, 0, np.minimum(1, d + v * x))

            def g_inv(x):
                return np.where(x <= d, 0, (x - d) / v)

            def g_prime(x):
                return v

        elif self._name == 'lep':
            # leverage equivalent pricing
            # self.r0 = risk free/financing and r = risk charge (the solved parameter)
            r = self.shape
            delta = r / (1 + r)
            d = self.r0 / (1 + self.r0)
            spread = delta - d
            self.has_mass = (d > 0)
            self.mass = d

            def g(x):
                return np.where(x == 0, 0, np.minimum(1, d + (1 - d) * x + spread * np.sqrt(x * (1 - x))))

            spread2 = spread ** 2
            a = (1 - d) ** 2 + spread2

            def g_inv(y):
                mb = (2 * (y - d) * (1 - d) + spread2)  # mb = -b
                c = (y - d) ** 2
                rad = np.sqrt(mb * mb - 4 * a * c)
                # l = (mb + rad)/(2 * a)
                u = (mb - rad) / (2 * a)
                return np.where(y < d, 0, np.maximum(0, u))

        elif self._name == 'dual':
            # dual moment
            p = self.shape
            q = 1 / p
            self.has_mass = False

            def g(x):
                return 1 - (1 - x)**p

            def g_inv(y):
                return 1 - (1 - y)**q

            def g_prime(x):
                return p * (1 - x)**(p - 1)

        elif self._name == 'wtdtvar' or self._name == 'bitvar':
            # weighted tvar, df = p0 <p1, shape = weight on p1
            try:
                p0, p1 = df
                assert p0 < p1
                w = shape
                self.has_mass = (p1 == 1)
                self.mass = w if p1 == 1 else 0
                pt = (1 - p1) / (1 - p0) * (1 - w) + w
                s = np.array([0.,  1-p1, 1-p0, 1.])
                gs = np.array([0.,   pt,   1., 1.])
                g = interp1d(s, gs, kind='linear')
                g_inv = interp1d(gs, s, kind='linear')

                if p1 < 1:
                    def g_prime(x):
                        return np.where(x > 1 - p0, 0,
                                        np.where(x < 1 - p1,
                                                 w / (1 - p1) + (1 - w) / (1 - p0),
                                                 (1 - w) / (1 - p0)))
                else:
                    # p1==1, don't want 1/(1-p1)
                    def g_prime(x):
                        return np.where(x > 1 - p0, 0, (1 - w) / (1 - p0))

            except:
                raise ValueError('Inadmissible parameters to Distortion for wtdtvar. '
                                 'Pass shape=wt for p1 and df=[p0, p1]')

        elif self._name == 'convex':
            # convex envelope and general interpolation
            # NOT ALLOWED to have a mass...
            self.has_mass = False
            # use shape for number of points in calibrating data set
            self.shape = f'on {len(df):d} points'
            if not (0 in df[col_x].values and 1 in df[col_x].values):
                # painful...always want 0 and 1 there...but don't know what other columns in df
                # logger.debug('df does not contain s=0/1...adding')
                df = df[[col_x, col_y]].copy().reset_index(drop=True)
                df.loc[len(df)] = (0,0)
                df.loc[len(df)] = (1,1)
                df = df.sort_values(col_x)
            if len(df) > 2:
                hull = ConvexHull(df[[col_x, col_y]])
                knots = list(set(hull.simplices.flatten()))
                g = interp1d(df.iloc[knots, df.columns.get_loc(col_x)],
                             df.iloc[knots, df.columns.get_loc(col_y)], kind='linear',
                             bounds_error=False, fill_value=(0,1))
                g_inv = interp1d(df.iloc[knots, df.columns.get_loc(col_y)],
                             df.iloc[knots, df.columns.get_loc(col_x)], kind='linear',
                             bounds_error=False, fill_value=(0,1))
            else:
                df = df.sort_values(col_x)
                g = interp1d(df[col_x], df[col_y], kind='linear',
                             bounds_error=False, fill_value=(0,1))
                g_inv = interp1d(df[col_y], df[col_x], kind='linear',
                             bounds_error=False, fill_value=(0,1))
        else:
            raise ValueError(
                "Incorrect spec passed to Distortion; implemented g types are ph, wang, tvar, "
                "ly (linear yield), lep (layer equivalent pricing) and clin (clipped linear)")

        self.g = g
        self.g_inv = g_inv
        if g_prime is None:
            g_prime = lambda x: (g(x + 1e-6) - g(x - 1e-6)) / 2e-6
        self.g_prime = g_prime

    def g_dual(self, x):
        """
        The dual of the distortion function g.
        """
        return 1 - self.g(1 - x)

    def __str__(self):
        if self.display_name != '':
            s = self.display_name
            return s
        elif isinstance(self.shape, str):
            s = f'{self._distortion_names_.get(self._name, self._name)}, {self.shape}'
        else:
            s = f'{self._distortion_names_.get(self._name, self._name)}, {self.shape:.3f}'
        if self._name == 'tt':
            s += f', {self.df:.2f}'
        if self._name == 'wtdtvar':
            s += f', ({self.df[0]:.3f}/{self.df[1]:.3f})'
        elif self.has_mass:
            # don't show the mass for wtdtvar
            s += f', mass {self.mass:.3f}'

        return s

    def __repr__(self):
        s = f'{self.name} ({self.shape}'
        if self.has_mass:
            s += f', {self.r0})'
        elif self._name == 'tt':
            s += f', {self.df:.2f})'
        else:
            s += ')'
        return s

    @property
    def name(self):
        return self.display_name if self.display_name != '' else self._name

    @name.setter
    def name(self, value):
        self._name = value

    def plot(self, xs=None, n=101, both=True, ax=None, plot_points=True, scale='linear', c=None, **kwargs):
        """
        Quick plot of the distortion

        :param xs:
        :param n:  length of vector is no xs
        :param both: True: plot g and ginv and add decorations, if False just g and no trimmings
        :param ax:
        :param plot_points:
        :param scale: linear as usual or return plots -log(gs)  vs -logs and inverts both scales
        :param kwargs:  passed to plot
        :return:
        """

        assert scale in ['linear', 'return']

        if scale == 'return' and n == 101:
            # default not enough for return
            n = 10001

        if xs is None:
            xs = np.linspace(0, 1, n)

        y1 = self.g(xs)
        if both:
            y2 = self.g_inv(xs)

        if ax is None:
            fig, ax = plt.subplots(1,1, figsize=(FIG_H, FIG_H), constrained_layout=True)

        if scale == 'linear':
            ax.plot(xs, y1, c='C0', label='$g$', **kwargs)
            if both:
                ax.plot(xs, y2, c='C1', label='$g^{-1}$', **kwargs)
            ax.plot(xs, xs, lw=0.5, color='black', alpha=0.5)
        elif scale == 'return':
            ax.plot(xs, y1, c='C0', label='$g$', **kwargs)
            if both:
                ax.plot(xs, y2, c='C1', label='$g^{-1}$', **kwargs)
            ax.set(xscale='log', yscale='log', xlim=[1/5000, 1], ylim=[1/5000, 1])
            ax.plot(xs, xs, lw=0.5, color='black', alpha=0.5)

        if self._name == 'convex' and plot_points:
            if len(self.df) > 50:
                alpha = .35
            elif len(self.df) > 20:
                alpha = 0.6
            else:
                alpha = 1
            if c is None:
                c = 'C2'
            if scale == 'linear':
                ax.scatter(x=self.df[self.col_x], y=self.df[self.col_y], marker='.', s=15, color=c, alpha=alpha)
            elif scale == 'return':
                ax.scatter(x=1/self.df[self.col_x], y=1/self.df[self.col_y], marker='.', s=15, color=c, alpha=alpha)

        ax.set(title=fill(str(self), 20), aspect='equal');

        return ax

    @classmethod
    def test(cls, r0=0.035, df=[0.0, .9]):
        """
        Tester: make some nice plots of available distortions.

        :return:
        """

        f0, axs0 = plt.subplots(2, 11, figsize=(22, 4), constrained_layout=True, sharex=True, sharey=True)
        f1, axs1 = plt.subplots(2, 11, figsize=(22, 4), constrained_layout=True, sharex=True, sharey=True)
        axiter0 = iter(axs0.flat)
        axiter1 = iter(axs1.flat)

        xs = np.linspace(0, 1, 1001)

        # zip stops at the shorter of the vectors, so this does not include convex (must be listed last)
        # added df for the t; everyone else can ignore it
        # rank by order on large lsoses...
        for axiter, scale in zip([axiter0, axiter1], ['linear', 'return']):
            for name, shape in zip(cls._available_distortions_, cls._eg_param_1_):
                dist = Distortion(name, shape, r0, df=df)
                dist.plot(xs, ax=next(axiter), scale=scale)

            dist = Distortion.convex_example('bond')
            dist.plot(xs, ax=next(axiter), scale=scale)

            # order will look better like this
            for name, shape in zip(cls._available_distortions_, cls._eg_param_2_):
                dist = Distortion(name, shape, r0, df=df)
                dist.plot(xs, ax=next(axiter), scale=scale)

            dist = Distortion.convex_example('cat')
            dist.plot(xs, ax=next(axiter), scale=scale)

        # tidy up
        for ax in axiter0:
            f0.delaxes(ax)
        for ax in axiter1:
            f1.delaxes(ax)

        f0.suptitle('Example Distortion Functions - Linear Scale')
        f1.suptitle('Example Distortion Functions - Return Scale')

    @staticmethod
    def distortions_from_params(params, index, r0=0.025, df=5.5, pricing=True, strict=True):
        """
        Make set of dist funs and inverses from params, output of port.calibrate_distortions.
        params must just have one row for each method and be in the output format of cal_dist.

        Called by Portfolio.

        :param index:
        :param params: dataframe such that params[index, :] has a [lep, param] etc.
               pricing=True, strict=True: which distortions to allow
               df for t distribution
        :param r0: min rol parameters
        :param strict:
        :param pricing:
        :return:
        """

        temp = params.loc[index, :]
        dists = {}
        for dn in Distortion.available_distortions(pricing=pricing, strict=strict):
            param = float(temp.loc[dn, 'param'])
            dists[dn] = Distortion(name=dn, shape=param, r0=r0, df=df)

        return dists

    @staticmethod
    def convex_example(source='bond'):
        """
        Example convex distortion using data from https://www.bis.org/publ/qtrpdf/r_qt0312e.pdf.

        :param source: bond gives a bond yield curve example, cat gives cat bond / cat reinsurance pricing based example
        :return:
        """

        if source == 'bond':
            yield_curve = '''
            AAAA    0.000000  0.000000
            AAA     0.000018  0.006386
            AA      0.000144  0.007122
            A       0.000278  0.010291
            BBB     0.002012  0.017089
            BB      0.012674  0.036455
            B       0.040052  0.069181
            Z       1.000000  1.000000'''

            df = pd.read_fwf(StringIO(yield_curve))
            df.columns = ['Rating', 'EL', 'Yield']
            return Distortion('convex', 'Yield Curve', df=df, col_x='EL', col_y='Yield')

        elif source.lower() == 'cat':
            cat_bond = '''EL,ROL
            0.116196,0.32613
            0.088113,0.2452
            0.074811,0.22769
            0.056385,0.17131
            0.046923,0.15326
            0.032961,0.12222
            0.02807,0.11037
            0.024205,0.1022
            0.011564,0.07284
            0.005813,0.06004
            0,0
            1,1'''
            df = pd.read_csv(StringIO(cat_bond))
            return Distortion('convex', 'Cat Bond', df=df, col_x='EL', col_y='ROL')

        else:
            raise ValueError(f'Inadmissible value {source} passed to convex_example, expected yield or cat')

    @staticmethod
    def bagged_distortion(data, proportion, samples, display_name="", random_state=None):
        """
        Make a distortion by bootstrap aggregation (Bagging) resampling, taking the convex envelope,
        and averaging from data.

        Each sample uses proportion of the data.

        Data must have  two columns: EL and Spread

        :param data:
        :param proportion: proportion of data for each sample
        :param samples: number of resamples
        :param display_name: display_name of created distortion
        :param random_state: for pd.sample....ensures reproducibility
        :return:
        """

        df = pd.DataFrame(index=np.linspace(0,1,10001), dtype=np.float)

        for i in range(samples):
            rebit = data.sample(frac=proportion, replace=False, random_state=random_state)
            rebit.loc[-1] = [0, 0]
            rebit.loc[max(rebit.index)+1] = [1, 1]
            d = Distortion('convex', 0, df=rebit, col_x='EL', col_y='Spread')
            df[i] = d.g(df.index)

        df['avg'] = df.mean(axis=1)
        df2 =df['avg'].copy()
        df2.index.name = 's'
        df2 = df2.reset_index(drop=False)

        d = Distortion('convex', 0, df=df2, col_x='s', col_y='avg', display_name=display_name)

        return d

    @staticmethod
    def average_distortion(data, display_name, n=201, el_col='EL', spread_col='Spread'):
        """
        Create average distortion from (s, g(s)) pairs. Each point defines a wtdTVaR with
        p=s and p=1 points.

        :param data:
        :param display_name:
        :param n: number of s values (between 0 and max(EL), 1 is added
        :param el_col:   column containing EL
        :param spread_col: column containing Spread
        :return:
        """

        els = data[el_col]
        spreads = data[spread_col]
        max_el = els.max()
        s = np.hstack((np.linspace(0, max_el, n), 1))
        ans = np.zeros((len(s), len(data)))
        for i, el, spread in zip(range(len(data)), els, spreads):
            p = 1 - el
            w = (spread - el) / (1 - el)
            d = Distortion('wtdtvar', w, df=[0,p])
            ans[:, i] = d.g(s)

        df = pd.DataFrame({'s': s, 'gs': np.mean(ans, 1)})
        dout = Distortion('convex', None, df=df, col_x='s', col_y='gs', display_name=display_name)
        return dout

    @staticmethod
    def wtd_tvar(ps, wts, display_name='', details=False):
        """
        A careful version of wtd tvar with knots at ps and wts.

        :param ps:
        :param wts:
        :param display_name:
        :param details:
        :return:
        """

        # evaluate at 0, 1 and all the knot points
        ps0 = np.array(ps)
        s = np.array(sorted(set((0.,1.)).union(1-ps0)))
        s = s.reshape(len(s), 1)

        wts = np.array(wts).reshape(len(wts), 1)
        if np.sum(wts) != 1:
            wts = wts / np.sum(wts)
        ps = np.array(ps).reshape(1, len(ps))

        gs = np.where(ps == 1, 1, np.minimum(s / (1 - ps), 1)) @ wts

        d = Distortion.s_gs_distortion(s, gs, display_name)
        if details:
            return d, s, gs
        else:
            return d

    @staticmethod
    def s_gs_distortion(s, gs, display_name=''):
        """
        Make a convex envelope distortion from {s, g(s)} points.

        :param s: iterable (can be converted into numpy.array
        :param gs:
        :param display_name:
        :return:
        """
        s = np.array(s)
        gs = np.array(gs)
        return Distortion('convex', None, df=pd.DataFrame({'s': s.flat, 'gs': gs.flat}),
                          col_x='s', col_y='gs', display_name=display_name)

    def price(self, ser, a=np.inf, kind='ask', S_calculation='forwards'):
        r"""
        Compute the bid and ask prices for the distribution determined by ``ser`` with
        an asset limit ``a``. Index of ``ser`` need not be equally spaced, so it can
        be applied to :math:`\kappa`. To do this for unit A in portfolio port::

            ser = port.density_df[['exeqa_A', 'p_total']].\
                set_index('exeqa_A').groupby('exeqa_A').\
                sum()['p_total']
            dist.price(ser, port.q(0.99), 'both')

        Always use ``S_calculation='forwards`` method to compute S = 1 - cumsum(probs).
        Computes the price as the integral of gS.

        :param ser: pd.Series of is probabilities, indexed by outcomes. Outcomes need not
          be spaced evenly. ``ser`` is usually a probability column from ``density_df``.
        :param kind: is "ask", "bid",  or "both", giving the pricing view.
        :param a: asset level. ``ser`` is truncated at ``a``.
        """

        if not isinstance(ser, pd.Series):
            raise ValueError(f'ser must be a pandas Series, not {type(ser)}')

        if kind in ['bid', 'ask']:
            pass
        elif kind == 'both':
            return [*self.price(ser, a, 'bid', S_calculation),
                    self.price(ser, a, 'ask', S_calculation)[0]]
        else:
            raise ValueError(f'kind must be bid, ask, or both, not {kind}')

        # apply limit
        if a < np.inf:
            # ser = ser.copy()
            # tail = ser[a:].sum()
            ser = ser[:a]
            # ser[a] = tail

        # unlikely to be an issue
        ser = ser.sort_index(ascending=True)

        if S_calculation == 'forwards':
            S = 1 - ser.cumsum()
            S = np.maximum(0, S)
        else:
            fill_value = max(0, 1 - ser.sum())
            S = np.minimum(1, fill_value +
                           ser.shift(-1, fill_value=0)[::-1].cumsum()[::-1])

        # not all distortions return numpy; force conversion
        if kind == 'ask':
            gS = np.array(self.g(S))
        else:
            gS = np.array(self.g_dual(S))

        # trapz is not correct with our interpretation elsewhere; we use a step function
        # loss = np.trapz(S, S.index)
        # prem = np.trapz(gS, S.index)
        dx = np.diff(S.index)
        loss = (S.iloc[:-1].values * dx).sum()
        prem = (gS[:-1] * dx).sum()
        return prem, loss

    def price2(self, ser, a=None, S_calculation='forwards'):
        r"""
        Compute the bid and ask prices for the distribution determined by ``ser`` with
        an asset limits given by values of ``ser``. Index of ``ser`` need not be equally
        spaced, so it can be applied to :math:`\kappa`. To do this for unit A in portfolio
        ``port``::

            ser = port.density_df[['exeqa_A', 'p_total']].\
                set_index('exeqa_A').groupby('exeqa_A').\
                sum()['p_total']
            dist.price(ser, port.q(0.99))

        :param ser: pd.Series of is probabilities, indexed by outcomes. Outcomes must
          be spaced evenly. ``ser`` is usually a probability column from ``density_df``.
        """

        if not isinstance(ser, pd.Series):
            raise ValueError(f'ser must be a pandas Series, not {type(ser)}')

        if S_calculation == 'forwards':
            S = 1 - ser.cumsum()
            S = np.maximum(0, S)
        else:
            fill_value = max(0, 1 - ser.sum())
            S = np.minimum(1, fill_value +
                           ser.shift(-1, fill_value=0)[::-1].cumsum()[::-1])

        # not all distortions return numpy; force conversion
        gS = np.array(self.g(S))
        dgS = np.array(self.g_dual(S))

        dx = np.diff(S.index)
        loss = (S.iloc[:-1].values * dx).cumsum()
        ask = (gS[:-1] * dx).cumsum()
        bid = (dgS[:-1] * dx).cumsum()
        # index is shifted by 1 because it is the right hand range of integration
        ans = pd.DataFrame({'bid': bid, 'el': loss, 'ask': ask}, index=S.index[1:])
        if a is None:
            return ans
        else:
            # no longer guaranteed that a is in ser.index
            return ans.iloc[ans.index.get_loc(a, method='nearest')]


# make the hu-scs case study

from aggregate.extensions import case_studies as cs
from aggregate import build
import numpy as np

if __name__ == '__main__':

      # parameters from PIR
      freq1, sigma1 = 70, 1.9
      freq2, sigma2 = 2, 2.5
      mu1 = -sigma1**2 / 2
      mu2 = -sigma2**2 / 2
      sev1, sev2 = 1, 15
      print(mu1, sigma1, mu2, sigma2,
            freq1, np.exp(mu1), sigma1,
            freq2, np.exp(mu2), sigma2)

      recalc = build(f'sev Husev {15 * np.exp(mu2)} * lognorm {sigma2}')
      a, d = recalc.isf(0.05), recalc.isf(0.005)
      y = d - a
      print(y, a, d)

      # create exhibits
      hs = cs.CaseStudy()
      hs.factory(case_id='hs',
                 case_name='Hu/SCS Case',
                 case_description='Hu/SCS Case in the new syntax.',
                 a_distribution=f'agg SCS {freq1} claims sev {sev1 * np.exp(mu1)} * lognorm {sigma1} poisson',
                 b_distribution_gross=f'agg Hu {freq2} claims sev {sev2 * np.exp(mu2)} * lognorm {sigma2} poisson',
                 b_distribution_net=f'agg Hu {freq2} claims sev {sev2 * np.exp(mu2)} * lognorm {sigma2} poisson ' \
                                    f'aggregate net of {y} xs {a}',
                 reg_p=0.999,
                 roe=0.10,
                 d2tc=0.3,
                 f_discrete=False,
                 s_values=[.005, 0.01, 0.03],
                 gs_values=[0.029126,   0.047619,   0.074074],
                 bs=1/4,
                 log2=19,
                 padding=1)
      hs.full_monty()
      hs.to_json()
      hs.browse_exhibits()


      hs2 = cs.CaseStudy()
      hs2.factory(case_id='hs_per_occ',
                  case_name='Hu/SCS',
                  case_description='Hu/SCS Case in the new syntax with per occurrence reinsurance .',
                  a_distribution=f'agg SCS {freq1} claims sev {sev1 * np.exp(mu1)}  * lognorm {sigma1} poisson',
                  b_distribution_gross=f'agg Hu {freq2} claims sev {sev2 * np.exp(mu2)} * lognorm {sigma2} poisson',
                  b_distribution_net=f'agg Hu {freq2} claims sev {sev2 * np.exp(mu2)} * lognorm {sigma2} '
                                     f'occurrence net of {y} xs {a} poisson',
                  reg_p=0.999,
                  roe=0.10,
                  d2tc=0.3,
                  f_discrete=False,
                  s_values=[.005, 0.01, 0.03],
                  gs_values=[0.029126,   0.047619,   0.074074],
                  bs=1/4,
                  log2=19,
                  padding=1)
      hs2.full_monty()
      hs2.to_json()



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.fft import irfft, fft, ifft, rfft, fftshift
from numpy import real, imag, roll
from .. import build
from .. constants import FIG_H, FIG_W
import matplotlib as mpl
from aggregate import build, qd

def fft_wrapping_illustration(ez=10, en=20, sev_clause='', small2=0):
    """
    Illustrate wrapping by convolving a uniform distribution with mean ez
    en times (if ``ez>0`` or ``sev_clause!=''``) or using the input ``sev_clause``.
    ``sev_clause`` should be a ``dsev`` tailored to ``bs==1``.

    Show in a space just big enough for the severity first and then
    big enough for the full aggregate. Center and right hand plot illustrate how the
    full components are sliced up and combined to the wrapped total.

    If small2 is zero it is taken to be the smallest value to "fit" the severity.

    (moved from figures.py)

    """
    fig, axs = plt.subplots(1, 3, figsize=(3 * FIG_W, FIG_H), constrained_layout=True)
    ax0, ax1, ax2 = axs.flat

    if ez == 0 or sev_clause != '':
        if sev_clause == '':
            raise ValueError('Must input one of ez>0 or a valid DecL sev_clause')
        sev = build(f'agg Junk 1 claim {sev_clause} fixed')
        ez = sev.sev_m
        q1 = sev.q(1)
        if small2 == 0:
            small2 = int(np.ceil(np.log2(q1)))
        xs = np.hstack((-np.inf, np.arange(1 << small2) + 0.5))
        z = np.diff(sev.sev.cdf(xs))
        # enough space for aggregate
        big2 = int(np.ceil(np.log2(q1 * en)))
    else:
        # enough space for severity and make sev
        if small2 == 0:
            small2 = int(np.ceil(np.log2(2 * ez)))
        z = np.zeros(1 << small2)
        z[:ez*2] = 1 / ez / 2
        # enough space for aggregate
        big2 = int(np.ceil(np.log2(2 * ez * en)))

    if big2 <= 8:
        ds = 'steps-post'
    else:
        ds = 'default'

    wrapped = irfft( rfft(z) ** en )
    ax0.plot(wrapped, c='C0', drawstyle=ds)
    ax0.set(title=f'Wrapped distribution\nlog2={small2}')
    lm = ax0.get_ylim()
    lm = (-lm[1] / 20, lm[1]* 1.1)
    ax0.set(ylim=lm)

    full = irfft( rfft(z, 1<<big2) ** en )
    norm = mpl.colors.Normalize(0, 1, clip=True)
    cmappable = mpl.cm.ScalarMappable(norm=norm, cmap='plasma')
    mapper = cmappable.to_rgba
    cc = list(map(mapper, np.linspace(0, 1, 1 << big2-small2)))
    ax1.plot(full, label='Full computation', c='w', alpha=1, lw=3, drawstyle=ds)
    for n, (s, c) in enumerate(zip(full.reshape((1<<big2-small2, 1<<small2)), cc)):
        ax1.plot(s, c=c, label=f'Part {n}', drawstyle=ds)
        ax1.plot(np.arange((1<<small2) * n, (1<<small2) * (n+1)), s, c=c, lw=2,
                 drawstyle=ds, label=None)
    for n in range(1 << big2-small2):
        ax1.axvline(n * (1 << small2), lw=.25, c='C7')
    for n in range(1 << big2-small2):
        ax1.axvline(n * (1 << small2), lw=.25, c='C7')
    if big2 - small2 <= 3:
        ax1.legend(loc='center right')
    ax1.set(title=f'Full distribution\nlog2={big2}, {1<<big2-small2} components')

    wrapped_from_full = full.reshape((1<<big2-small2, 1<<small2))
    ax2.plot(wrapped_from_full.T, label=None, c='C7', lw=.5, drawstyle=ds)
    ax2.plot(wrapped_from_full.sum(0), lw=3
             , drawstyle=ds, label='Wrapped from full', c='C1')
    ax2.plot(wrapped, lw=1, label='Wrapped', c='C0', drawstyle=ds)
    ax2.set(title='Wrapping components (grey)\nSums (blue, organge as left)')
    # ax2.legend(loc)
    ax2.set(ylim=lm)

    assert np.allclose(wrapped_from_full.sum(0), wrapped)


def ft_invert(log2, chf, frz_generator, params, loc=0, scale=1, xmax=0, xshift=0,
              suptitle='', wraps=None, disc_calc='density'):
    """
    Illustrate "manual" inversion of a characteristic function using irfft, including
    optional scaling and location shift.


    :param params: a list of shape parameters. loc and scale are handled separately.
    :param chf: the characteristic function of the distribution, takes args params, loc, scale, t;
      routine handles conversion to Fourier Transform
    :param frz_generator: the scipy.stats function to create the underlying distribution.
      Used to compute the exact answer. If there is not analytic formula, you can pass in
      a numpy array with the values of the distribution at the points in xs computed by
      some other means (e.g., for the Tweedie distribution).
    :param loc: location paramteter
    :param scale: scale parameter
    :param xmax: if not zero, used to fix xmax. Otherwise, selected as frz.isf(1e-17) to capture the
      full range of theunderlying distribution. For thick tailed distributions, you usually input
      xmax manually. Note bs / xmax / n. If None set equal to n = 1 << log2, for use in
      discrete distributions (forced bucket size of 1).
    :param xshift: if not zero, used to shift the x-axis. The minimum x value equals xshift.
      To center the x-axis, set xshift = -xmax/2. Should be a multiple of bs = xmax / n.
    :param suptitle: optional suptitle for the figure
    :param wraps: optional list of wrap values.
    :param disc_calc: 'density' rescales pdf, 'surival' uses backward differences of sf (to
      match Aggregate class calculation)
    """

    # number of buckets
    n = 1 << log2
    if xmax is None:
        xmax = n

    # make frozen object
    if callable(frz_generator):
        if scale is None:
            # freq dists (Poisson) do not allow scaling
            frz = frz_generator(*params, loc=loc)
            frz.pdf = frz.pmf
            # for subsequent use
            scale = 1
        else:
            frz = frz_generator(*params, loc=loc, scale=scale)

    # spatial upto xmax; used to create exact using the scipy stats object
    # sampling interval (wavelength) = xmax / n
    if xmax == 0:
        xmax = frz.isf(1e-17)
    # sampling domain, for exact and to "label" the Fourier Transform output
    # xs = np.arange(n) * xmax / n
    bs = xmax / n
    xs = np.arange(n) * bs + xshift
    if callable(frz_generator):
        if disc_calc == 'density':
            exact = frz.pdf(xs)
            exact = exact / exact.sum() * (frz.cdf(xs[-1]) - frz.cdf(xs[0]))
        else:
            xs1 = np.hstack((xs - bs / 2, xs[-1] + bs / 2))
            exact = -np.diff(frz.sf(xs1))
    else:
        # pass in values
        exact = frz_generator

    # convert chf to ft including scale and loc effects
    def loc_ft(t):
        nonlocal params, loc, scale
        # ft(t)= int f(x)exp(-2πi t x)dx
        # chf(t) = int f(x)exp(i t x)dx
        t1 = -t * 2 * np.pi
        ans = chf(*params, t1 * scale)
        if loc != 0:
            # for some reason ans *= np.exp(-t1 * loc) does not work
            ans = ans * np.exp(t1 * loc * 1j)
        return ans

    # sampling interval = bs = xmax / n [small bs, high sampling rate]
    # sampling freq is 1 / bs = n / xmax, the highest sampling freq for inverting the FT
    # note xmax = n * bs, so n / xmax = 1 / bs.
    # f(x) = int_R fhat(t) exp(2πi tx)dt ≈ int_-f_max_f^max_f ...
    f_max = n / xmax
    # sample the FT; using real fft, only need half the range
    ts = np.arange(n // 2 + 1) * f_max / n   # ts = np.arange(n // 2 + 1) / xmax
    fx = loc_ft(ts)
    # for debugging
    ft_invert.fx = fx
    ft_invert.ts = ts
    x = irfft(fx)
    if xshift != 0:
        x = np.roll(x, -int(xshift / bs))

    # plotting
    fig, axs = plt.subplots(2, 2, figsize=(2 * 3.5, 2 * 2.45), constrained_layout=True)
    ax0, ax1, ax2, ax3 = axs.flat

    for ax in axs[0].flat:
        ax.plot(xs, exact, label='exact', lw=1.5)
        ax.plot(xs, x, label='xs irfft', ls=':', lw=1.5)
        ax.legend(fontsize='x-small')
    ax0.set(title='Density', xlabel='Outcome, x')
    # mn = min(np.log10(exact).min(), np.log10(x).min())
    mn0 = np.log10(x).min() * 1.25
    mn = 10 ** np.floor(mn0)
    mx = max(np.log10(exact).max(), np.log10(x).max())
    mx = 10 ** np.ceil(mx)
    if np.isnan(mn):
        mn = 1e-17
    if np.isnan(mx):
        mx = 1
    ax1.set(yscale='log', ylim=[mn, mx], title='Log density', xlabel='Outcome, x')

    # amplitude and phase
    ax2.plot(ts, np.abs(fx), '-', lw=1.5, c='C3')
    ax2.set(title='Amplitude',
            ylabel='|ft|', yscale='log', xlabel='frequency')
    if log2 <= 8:
        ax3.plot(ts, np.cumsum(np.angle(fx)) / (2 * np.pi), '-', marker='.', ms=3, c='C2')
    else:
        ax3.plot(ts, np.cumsum(np.angle(fx)) / (2 * np.pi), c='C2')
    ax3.set(title='Cumulative phase',
            ylabel='cumsum(arg(ft)) / $2\pi$', xlabel='frequency')

    if suptitle != '':
        fig.suptitle(suptitle)

    if wraps is not None:
        fig2, ax = plt.subplots(1, 1, figsize=(3.5 * 2, 2.45 * 2), constrained_layout=True)
        rt = exact.copy()
        ax.plot(xs, exact, label='exact', lw=1.5)
        ax.plot(xs, x, label='xs irfft', ls=':', lw=1.5)
        for b in wraps:
            xs2 = b * n / f_max + xs
            adj = frz.pdf(xs2)
            adj = adj / np.sum(adj) * (frz.cdf((b + 1) * n / f_max) - frz.cdf(b * n / f_max))
            rt += adj
            ax.plot(xs, rt, label=f'wrap {b}', lw=.5)
        ax.set(yscale='log', ylim=[mn, mx], title='Aliasing analysis',
               xlabel='Outcome, x', ylabel='Log density')
        ax.legend(fontsize='x-small', ncol=2, loc='upper right')

    return pd.DataFrame({'x': xs, 'p': x, 'p_exact': exact}).set_index('x')


def recentering_convolution_example(sev_clause, en, log2, agg_log2=0, bs=1,
                                    freq_clause='poisson', remove_fuzz=False):
    """
    Illustrate how to find the "correct" part of an aggregate and
    recenter it appropriately. Aggregate is::

        agg RecenteringExample en claims dsev xs ps poisson

    **Method**

    #. Compute aggregate with enough space for the supported part of the
      distribution (say, where density > 1e-15 or so)
    #. Subtract the mean by rolling left (negative shift) by mean / bs buckets (mod n)
    #. fft shift = roll (in either direction) by n / 2 buckets, because
      the distribution is centered at zero and has positive and negative parts.
    #. Set appropriate x values to align with the density. Density is from
      mean - n/2 to mean + n/2 - 1 (times bs).

    Reasonable defaults::

        en = 5000
        log2 = 15

        xs = [3, 4, 7, 34]
        ps = [1/8, 1/8, 1/8, 5/8]

    (from hifreq.py)

    :param xs: array of x values for dsev
    :param ps: array of x values for dsev

    """
    df, ag = recentering_convolution(sev_clause, freq_clause, en, log2, bs, remove_fuzz)

    # update the ag object if agg_log2
    if agg_log2 == 0:
        # just plot
        fig, ax0 = plt.subplots(1, 1, figsize=(
            3.5, 2.45), constrained_layout=True)

        df.plot(ax=ax0, c='C0')
        ax0.set(
            title=f'Shifted, centered, re-indexed aggregate\nlog2={log2} buckets')
        bit = None

    elif agg_log2 > 0:
        ag.update(approximation='exact', log2=agg_log2, bs=bs, padding=0)
        qd(ag)
        print('-'*80)
        # percentiles - help determining log2 needed for hi freq calculation
        qd(pd.Series([ag.agg_sd, ag.q(0.001), ag.q(0.999),
                      ag.q(0.999999) - ag.q(0.000001)],
                     index=['std dev', 'p001', 'p999', 'range']))
        print('-'*80)

        # merge and compare
        fig, axs = plt.subplots(2, 3, figsize=(
            3 * 3.5, 2 * 2.45), constrained_layout=True)
        ax0, ax1, ax2, ax3, ax4, ax5 = axs.flat

        df.plot(ax=ax0, c='C0')
        ax0.set(
            title=f'Shifted, centered, re-indexed aggregate\nlog2={log2} buckets')
        bit = pd.concat((ag.density_df.p_total, df), axis=1, join='inner')
        bit.plot(ax=ax1)
        ax1.lines[1].set(ls='--')
        ax1.legend(loc='upper right')
        ax1.set(
            title=f'Shifted vs. agg object\n'
                  f'Linear scale; agg object log2 = {ag.log2}')
        bit.plot(ax=ax2, logy=True)
        ax2.lines[1].set(ls='--')
        ax2.legend(loc='upper right')
        ax2.set(title='Shifted vs. agg object\nLog scale')

        abs_error = (bit.p_total - bit.a).abs()
        abs_error.plot(ax=ax4, c='C2', logy=True,
                       title=f'Abs error\n'
                       f'Max {abs_error.max():.5g}; '
                       f'Avg {abs_error.mean():.5g}')
        rel_error = abs_error / bit.p_total
        rel_error = rel_error.loc[bit.p_total > 1e-15]
        rel_error.plot(ax=ax5, c='C2', logy=True,
                       title=f'Rel error p_total > 1e-15\n'
                       f'Max {rel_error.max():.5g}; '
                       f'Avg {rel_error.mean():.5g}')
        ax3.remove()

    else:
        raise ValueError('log2 >= 0!')

    return df, ag, bit


def recentering_convolution(sev_clause, freq_clause, en, log2, bs, remove_fuzz):
    """
    Compute hifreq convol for sev_clause, a DecL severity statement.
    Must have only one component for now.
    Illustrates how to find the "correct" part of an aggregate and
    recenter it appropriately. Aggregate is::

        agg RecenteringExample en claims dsev xs ps poisson

    **Method**

    #. Compute aggregate with enough space for the supported part of the
      distribution (say, where density > 1e-15 or so)
    #. Subtract the mean by rolling left (negative shift) by mean / bs buckets (mod n)
    #. fft shift = roll (in either direction) by n / 2 buckets, because
      the distribution is centered at zero and has positive and negative parts.
    #. Set appropriate x values to align with the density. Density is from
      mean - n/2 to mean + n/2 - 1 (times bs).

    """

    ag = build(f'agg SevEg {en} claims {sev_clause} {freq_clause}', update=False)
    ez = ag.sev_m
    mean_bucket = int(round(ez * en / bs))

    ft_len = 1 << log2
    ag.xs = np.linspace(0, ft_len * bs, ft_len, endpoint=False)
    ag.log2 = log2
    ag.bs = bs
    ag.padding = 0
    z = ag.discretize('round', 'survival', True)
    assert len(z) == 1
    z = z[0]
    fz = rfft(z, ft_len)
    # aggregate by hand
    # fa = np.exp(en*(fz - 1))
    fa = ag.mgf(en, fz)
    a = irfft(fa)

    # center roll left by ez and 1 << log2-1, former
    # is the mean offset, the latter is fftshift
    a = roll(a, -(mean_bucket % ft_len) + (1 << log2 - 1))

    # set up aligned to the appropriate xs values
    df = pd.DataFrame(
        {'n': np.arange(mean_bucket - (1 << log2 - 1),
                        mean_bucket + (1 << log2 - 1)) * bs,
         'a': a}
    ).set_index('n')
    if remove_fuzz:
        # remove fuzz
        eps = np.finfo(float).eps
        df.loc[df.a.abs() < 2 * eps, 'a'] = 0
    qd(stats(df).T)
    print('-'*80)
    return df, ag


def stats(df):
    ans = {}
    ns = {}
    ns[1] = np.array(df.index)
    ns[2] = ns[1] * ns[1]
    ns[3] = ns[2] * ns[1]
    for c in df:
        ans[c] = [np.sum(ns[i] * df[c]) for i in [1, 2, 3]]
    df = pd.DataFrame(ans, index=[1, 2, 3])
    df.loc['var'] = df.loc[2] - df.loc[1] ** 2
    df.loc['sd'] = df.loc['var'] ** .5
    df.loc['cv'] = df.loc['sd'] / df.loc[1]
    df.loc['skew'] = (df.loc[3] - 3 * df.loc[2] * df.loc[1] +
                      2 * df.loc[1]**3) / df.loc['sd'] ** 3
    return df


# from collections import namedtuple
import numpy as np
import logging
import pandas as pd
from pathlib import Path
import re
from IPython.display import HTML, display
# from inspect import signature

from .constants import *
from .portfolio import Portfolio
from .distributions import Aggregate, Severity
from .spectral import Distortion
from .parser import UnderwritingLexer, UnderwritingParser
from .utilities import logger_level, round_bucket, Answer, LoggerManager, qd, show_fig

logger = logging.getLogger(__name__)


# rejected: immutable
# WriteAnswer = namedtuple('WriteAnswer', ['kind', 'name', 'spec', 'program', 'object'])


class Underwriter(object):
    """
    The ``Underwriter`` class manages the creation of Aggregate and Portfolio objects, and
    maintains a database of standard Severity (curves) and Aggregate (unit or line level) objects
    called the knowledge base.

    - Handles persistence to and from agg files
    - Is interface into program parser
    - Handles safe lookup from the knowledge for parser

    Objects have a kind and a name. The kind is one of 'sev', 'agg' or 'port'. The name is a string.
    They have a representation as a program. When the program is interpreted it produces a dictionary spec
    that can be used to create the object. The static method factory can create any object from the
    (kind, name, spec, program) quartet, though, strictly, program is not needed.

    The underwriter knowledge is stored in a dataframe indexed by kind and name with columns
    spec and program.
    """

    def __init__(self, name='Rory', databases=None, update=False, log2=10, debug=False):
        """
        Create an underwriter object. The underwriter is the interface to the knowledge base
        of the aggregate system. It is the interface to the parser and the interpreter, and
        to the database of curves, portfolios and aggregates.

        :param name: name of underwriter. Defaults to Rory, after Rory Cline, the best underwriter
            I know and a supporter of an analytic approach to underwriting.
        :param databases: name or list of database files to read in on creation. if None: nothing loaded; if
            'default' (installed) or 'site' (user, in ~/aggregate/databases) database \\*.agg files in default or site
            directory are loaded. If 'all' both default and site databases loaded. A string refers to a single database;
            an interable of strings is also valid. See `read_database` for search path.
        :param update: if True, update database files with new objects.
        :param log2: log2 of number of buckets in discrete representation.  10 is 1024 buckets.
        :param debug: if True, print debug messages.
        """

        self.name = name
        self.update = update
        if log2 <= 0:
            raise ValueError('log2 must be > 0. The number of buckets used equals 2**log2.')
        self.log2 = log2
        self.debug = debug
        self.lexer = UnderwritingLexer()
        self.parser = UnderwritingParser(self.safe_lookup, debug)
        # stop pyCharm complaining
        # knowledge - accounts and line known to the underwriter
        self._knowledge = pd.DataFrame(columns=['kind', 'name', 'spec', 'program'], dtype=object).set_index(
            ['kind', 'name'])

        if databases == 'all':
            databases = ['default', 'site']
        elif type(databases) == str:
            databases = [databases]

        # default_dir is installed by pip and contain installation files
        self.default_dir = Path(__file__).parent / 'agg'

        # site dir is in Users's home directory and stores their files
        self.site_dir = Path.home() / 'aggregate/databases'
        # check site dir exists
        self.site_dir.mkdir(parents=True, exist_ok=True)

        # case dir
        self.case_dir = Path.home() / 'aggregate/cases'
        # check case dir exists
        self.case_dir.mkdir(parents=True, exist_ok=True)

        self.template_dir = self.default_dir.parent / 'templates'
        self.template_dir.mkdir(parents=True, exist_ok=True)

        # make sure all database entries are stored:
        if databases is None:
            # nothing to do
            databases = []

        if 'default' in databases:
            # add all databases in default_dir
            databases.remove('default')
            for fn in self.default_dir.glob('*.agg'):
                self.read_database(fn)

        if 'site' in databases:
            # add all user databases
            databases.remove('site')
            databases += list(self.site_dir.glob('*.agg'))

        for fn in databases:
            self.read_database(fn)

        # ?! ensure description prints correctly. A bit cheaky.
        pd.set_option('display.max_colwidth', 100)

    def read_database(self, fn):
        """
        read database of curves, aggs and portfolios. These can live in the default directory
        that is part of the instalation or ~/aggregate/

        fn can be a string filename, with or without extension. A .agg extension is
        added if there is no suffix. Search path:

        * in the current dir
        * in site_dir (user)
        * in default_dir (installation)

        :param fn: database file name

        """

        p = Path(fn)
        if p.suffix == '':
            p = p.with_suffix('.agg')
        if p.exists():
            db_path = p
        elif (self.site_dir / p).exists():
            db_path = self.site_dir / fn
        elif (self.default_dir / p).exists():
            db_path = self.default_dir / p
        else:
            logger.error(f'Database {fn} not found. Ignoring.')
            return

        try:
            program = db_path.read_text(encoding='utf-8')
        except Exception as e:
            logger.error(f'Error reading requested database {db_path.name}. Ignoring.')
        else:
            # read in, parse, save to sev/agg/port dictionaries
            # throw away answer...not creating anything
            # get rid of cosmetic spaces, but keep newline tabs (2 or more spaces)
            program = re.sub('^  +', '\t', program, flags=re.MULTILINE)
            program = re.sub(' +', ' ', program)
            logger.info(f'Reading database {fn}...')
            n = len(self._knowledge)
            self.interpret_program(program)
            n = len(self._knowledge) - n
            logger.info(f'Database {fn} read into knowledge, adding {n} entries.')

    def __getitem__(self, item):
        """
        handles self[item]

        item = 'Name' for all objects called Name
        item = ("kind", "name") for object of kind kind called name

        subscriptable: try user portfolios, b/in portfolios, line, severity
        to access specifically use severity or line methods

        :param item:
        :return:
        """
        # much less fancy version:
        if not isinstance(item, (str, tuple)):
            raise ValueError(f'item must be a str (name of object) or tuple (kind, name), not {type(item)}.')

        try:
            if type(item) == str:
                # name == item, any type
                rows = self._knowledge.xs(item, axis=0, level=1, drop_level=False)
            elif type(item) == tuple:
                # return a dataframe row
                rows = self._knowledge.loc[[item]]
        except KeyError:
            raise KeyError(f'Item {item} not found.')
        except TypeError as e:
            # TODO fix this "TypeError: unhashable type: 'slice'"
            raise KeyError(f'getitem TypeError looking for {item}, {e}')
        else:
            if len(rows) == 1:
                kind, name, spec, program = rows.reset_index().iloc[0]
                return Answer(kind=kind, name=name, spec=spec, program=program, object=None)
            else:
                raise KeyError(f'Error: no unique object found matching {item}. Found {len(rows)} objects.')

    def __repr__(self):
        import aggregate
        s = []
        s.append(f'underwriter        {self.name}')
        s.append(f'knowledge          {len(self._knowledge)} programs')
        s.append(f'version            {aggregate.__version__}')
        s.append(f'update             {self.update}')
        for k in ['log2', 'update', 'debug']:
            s.append(f'{k:<19s}{getattr(self, k)}')
        sd = self.site_dir.resolve().relative_to(Path.home())
        sd = f'~/{sd}'
        dd = self.default_dir.resolve()
        try:
            dd = dd.relative_to(Path.home())
            dd = f'~/{dd}'
        except ValueError:
            dd = str(dd)
        s.append(f'site dir           {sd}')
        s.append(f'default dir        {dd}')
        s.append( '')
        s.append( 'help')
        s.append( 'build.knowledge    list of all programs')
        s.append( 'build.qlist(pat)   list programs matching pattern')
        s.append( 'build.show(name)   build and display name')
        return '\n'.join(s)

    # def _repr_html_(self):
    #     import aggregate
    #     s = [f'<p><h3>Underwriter {self.name}</h3>',
    #          f'Version {aggregate.__version__}. '
    #          f'Knowledge contains {len(self._knowledge)} programs. '
    #          'Run <code>build.knowledge</code> for a DataFrame listing by kind and name. '
    #          'Run <code>build.show(name)</code> for more details, <code>name</code>, '
    #          'accepts wildcards and regular expressions.'
    #          '</p>'
    #          # '<br>',
    #          # self.knowledge.to_html(),
    #          f'<p>Settings: '
    #          ]
    #     for k in ['log2', 'update', 'debug']:
    #         s.append(f'<span style="color: red;">{k}</span>: {getattr(self, k)}; ')
    #     return '\n'.join(s) + '</p>'

    def factory(self, answer):
        """
        Create object of kind from spec, a dictionary.
        Creating from uw obviously needs the uw, so this is NOT a staticmethod!

        :param answer: an Answer class with members kind, name, spec, and program
        :return: creates answer.object
        """

        kind, name, spec, program, obj = answer.values()

        if obj is not None:
            logger.error(f'Surprising: obj from Answer not None, type {type(obj)}. It will be overwritten.')

        if kind == 'agg':
            obj = Aggregate(**spec)
            obj.program = program
        elif kind == 'port':
            # Portfolio expects name, agg_list, uw
            # agg list is a list of spects that can be passed to Aggregate
            # need to drop the 'agg', name before the spec that gets returned
            # by the parser. Hence:
            agg_list = [k for i, j, k in spec['spec']]
            obj = Portfolio(name, agg_list, uw=self)
            obj.program = program
        elif kind == 'sev':
            if 'sev_wt' in spec and spec['sev_wt'] != 1:
                logger.log(WL,
                    f'Mixed severity cannot be created, returning spec. You had {spec["sev_wt"]}, expected 1')
                obj = None
            else:
                obj = Severity(**spec)
                # ? set outside...
                obj.program = program
        elif kind == 'distortion':
            obj = Distortion(**spec)
            # ? set outside
            obj.program = program
        else:
            ValueError(f'Cannot build {kind} objects')

        # update the object
        answer['object'] = obj
        return answer

    @property
    def knowledge(self):
        return self._knowledge.sort_index()[['program', 'spec']]

    def write(self, portfolio_program, log2=0, bs=0, update=None, **kwargs):
        """
        Write a natural language program. Write carries out the following steps.

        1. Read in the program and cleans it (e.g. punctuation, parens etc. are
           removed and ignored, replace ; with new line etc.)

        2. Parse line by line to create a dictionary definition of sev, agg or port objects.

        3. Replace sev.name, agg.name and port.name references with their objects.

        4. If update set, update all created objects.

        Sample input::

            port MY_PORTFOLIO
                agg Line1 20  loss 3 x 2 sev gamma 5 cv 0.30 mixed gamma 0.4
                agg Line2 10  claims 3 x 2 sevgamma 12 cv 0.30 mixed gamma 1.2
                agg Line 3100  premium at 0.4 3 x 2 sev 4 @ lognormal 3 cv 0.8 fixed 1

        The indents are required if each agg item appears on a new line.

        See parser for full language spec! See Aggregate class for many examples.

        :param log2:
        :param bs:
        :param portfolio_program:
        :param update: override class default
        :param kwargs: passed to object's update method if ``update==True``
        :return: single created object or dictionary name: object
        """

        # prepare for update
        # what / how to do; little awkward: to make easier for user have to strip named update args
        # out of kwargs
        if update is None:
            update = self.update

        if update is True and log2 == 0:
            log2 = self.log2

        # first see if portfolio_program refers to a built-in object
        try:
            # calls __getitem__
            answer = self[portfolio_program]
        except (LookupError, TypeError):
            logger.debug(f'underwriter.write | object not found, processing as a program.')
        else:
            logger.debug(f'underwriter.write | {answer.kind} object found.')
            answer = self.factory(answer)
            if update:
                answer.object.update(log2, bs, **kwargs)
            # rationalize return to be the same as parsed programs
            # TODO test this code
            return [answer]

        # if you fall through to here then the portfolio_program did not refer to a built-in object
        # run the program, get the interpreter return value, the irv, which contains kind/name->spec,program
        irv = self.interpret_program(portfolio_program)
        rv = []
        for answer in irv:
            # create objects and update if needed
            answer = self.factory(answer)
            if answer not in irv:
                logger.error('OK THAT FAILED' * 20)
            if answer.object is not None:
                # this can fail for named mixed severities, which can only
                # be created in context of an agg... that behaviour is
                # useful for named severities though... hence:
                if update:
                    update_method = getattr(answer.object, 'update', None)
                    if update_method is not None:
                        update_method(log2, bs, **kwargs)
            rv.append(answer)

        # report on what has been done
        if rv is None:
            logger.log(WL, f'Program did not contain any output')
        else:
            if len(rv):
                logger.info(f'Program created {len(rv)} objects.')

        # return created objects
        return rv

    def write_from_file(self, file_name, log2=0, bs=0, update=False, **kwargs):
        """
        Read program from file. Delegates to write.

        :param file_name:
        :param log2:
        :param bs:
        :param update:
        :param kwargs:
        :return:
        """
        portfolio_program = Path(file_name).read_text(encoding='utf-8')
        return self.write(portfolio_program, log2=log2, bs=bs, update=update, **kwargs)

    def interpret_program(self, portfolio_program):
        """
        Preprocess and then parse a program one line at a time. Each output is
        stored in the Underwriter's knowledge database. No objects are created.

        Error handling through parser.

        :param portfolio_program:
        :return:
        """

        # Preprocess ---------------------------------------------------------------------
        portfolio_program = self.lexer.preprocess(portfolio_program)

        # create return value list
        rv = []

        # Parse and Postprocess-----------------------------------------------------------
        # self.parser.reset()
        # program_line_dict = {}
        for program_line in portfolio_program:
            logger.debug(program_line)
            # preprocessor only returns lines of length > 0
            try:
                # parser returns the type, name, and spec of the object
                # this is where you can marry up with the program
                kind, name, spec = self.parser.parse(self.lexer.tokenize(program_line))
            except ValueError as e:
                if isinstance(e.args[0], str):
                    logger.error(e)
                    raise e
                else:
                    t = e.args[0].type
                    v = e.args[0].value
                    i = e.args[0].index
                    txt2 = program_line[0:i] + f'>>>' + program_line[i:]
                    logger.error(f'Parse error in input "{txt2}"\nValue {v} of type {t} not expected')
                    raise e
            else:
                # store in uw dictionary and create if needed
                logger.info(f'answer out: {kind} object {name} parsed successfully...adding to knowledge')
                self._knowledge.loc[(kind, name), :] = [spec, program_line]
                rv.append(Answer(kind=kind, name=name, spec=spec, program=program_line, object=None))

        return rv

    # @staticmethod
    # def add_defaults(dict_in, kind='agg'):
    #     """
    #     add default values to dict_inin. Leave existing values unchanged
    #     Used to output to a data frame, where you want all columns completed
    #
    #     :param dict_in:
    #     :param kind:
    #     :return:
    #     """
    #
    #     print('running add_defaults\n' * 10)
    #
    #     # use inspect to get the defaults
    #     # obtain signature
    #     sig = signature(Aggregate.__init__)
    #
    #     # self and name --> bound signature
    #     bs = sig.bind(None, '')
    #     bs.apply_defaults()
    #     # remove self
    #     bs.arguments.pop('self')
    #     defaults = bs.arguments
    #
    #     if kind == 'agg':
    #         defaults.update(dict_in)
    #
    #     elif kind == 'sev':
    #         for k, v in defaults.items():
    #             if k[0:3] == 'sev' and k not in dict_in and k != 'sev_wt':
    #                 dict_in[k] = v

    def safe_lookup(self, buildinid):
        """
        Lookup buildinid=kind.name in uw to find expected kind and merge safely into self.arg_dict.

        Different from getitem because it splits the item into kind and name and double
        checks you get the expected kind.

        :param buildinid:  a string in kind.name format
        :return:
        """

        # allow for sev.WC.1 name
        kind, *name = buildinid.split('.')
        name = '.'.join(name)
        try:
            # lookup in Underwriter
            answer = self[(kind, name)]
            found_kind, found_name, spec, program, _ = answer.values()
        except LookupError as e:
            logger.error(f'ERROR id {kind}.{name} not found in the knowledge.')
            raise e
        logger.debug(f'UnderwritingParser.safe_lookup | retrieved {kind}.{name} as type {found_kind}.{found_name}')
        if found_kind != kind:
            raise ValueError(f'Error: type of {name} is  {found_kind}, not expected {kind}')
        # don't want to pass back the original
        spec = spec.copy()
        return spec

    @staticmethod
    def logger_level(level):
        """
        Convenience function.

        :param level:
        :return:
        """
        # set logger_level for all aggregate loggers
        logger_level(level)

    def build(self, program, update=True, log2=0, bs=0, recommend_p=0.999, logger_level=None, **kwargs):
        """
        Convenience function to make work easy for the user. Intelligent auto updating.
        Detects discrete distributions and sets ``bs = 1``.

        ``build`` method sets loger level to 30 by default.

        ``__call__`` is set equal to ``build``.

        :param program:
        :param update: build's update
        :param log2: 0 is default: Estimate log2 for discrete and self.log2 for all others. Inupt value over-rides
            and cancels discrete computation (good for large discrete outcomes where bucket happens to be 1.)
        :param bs:
        :param logger_level: temporary log(ger) level for this build
        :param recommend_p: passed to recommend bucket functions. Increase (closer to 1) for thick tailed distributions.
        :param kwargs: passed to update, e.g., padding. Note force_severity=True is applied automatically
        :return: created object(s)
        """

        # automatically puts level back at the end
        if logger_level is not None:
            lm = LoggerManager(logger_level)

        # make stuff
        # write will return a dict with keys (kind, name) and value a WriteAnswer namedtuple
        rv = self.write(program, update=False, force_severity=True)

        if rv is None or len(rv) == 0:
            logger.log(WL, 'build produced no output')
            return None

        # in this loop bs_ and log2_ are the values actually used for each update;
        # they do not overwrite the input default values
        for answer in rv:
            if answer.object is None:
                # object not created
                logger.info(f'Object {answer.name} of kind {answer.kind} returned as '
                            'a spec; no further processing.')
            elif isinstance(answer.object, Aggregate) and update is True:
                # try to guess good defaults
                d = answer.spec
                if d['sev_name'] == 'dhistogram' and log2 == 0:
                    bs_ = 1
                    # how big?
                    if d['freq_name'] == 'fixed':
                        max_loss = np.max(d['sev_xs']) * d['exp_en']
                    elif d['freq_name'] == 'empirical':
                        max_loss = np.max(d['sev_xs']) * max(d['freq_a'])
                    elif d['freq_name'] == 'bernoulli':
                        # allow for max loss to occur
                        max_loss = np.max(d['sev_xs'])
                    else:
                        # normal approx on count
                        max_loss = np.max(d['sev_xs']) * d['exp_en'] * (1 + 3 * d['exp_en'] ** 0.5)
                    # binaries are 0b111... len-2 * 2 is len - 1
                    log2_ = len(bin(int(max_loss))) - 1
                    logger.info(f'({answer.kind}, {answer.name}): Discrete mode, '
                                'using bs=1 and log2={log2_}')
                else:
                    if log2 == 0:
                        log2_ = self.log2
                    else:
                        log2_ = log2
                    if bs == 0:
                        bs_ = round_bucket(answer.object.recommend_bucket(log2_, p=recommend_p))
                    else:
                        bs_ = bs
                    logger.info(f'({answer.kind}, {answer.name}): Normal mode, using bs={bs_} (1/{1/bs_}) and log2={log2_}')
                try:
                    answer.object.update(log2=log2_, bs=bs_, debug=self.debug, force_severity=True, **kwargs)
                except ZeroDivisionError as e:
                    logger.error(e)
                except AttributeError as e:
                    logger.error(e)
            elif isinstance(answer.object, Severity):
                # there is no updating for severities
                pass
            elif isinstance(answer.object, Portfolio) and update is True:
                # figure stuff
                if log2 == -1:
                    log2_ = 13
                elif log2 == 0:
                    log2_ = self.log2
                else:
                    log2_ = log2
                if bs == 0:
                    bs_ = answer.object.best_bucket(log2_)
                else:
                    bs_ = bs
                logger.info(f'updating with {log2}, bs=1/{1 / bs_}')
                logger.info(f'({answer.kind}, {answer.name}): bs={bs_} and log2={log2_}')
                answer.object.update(log2=log2_, bs=bs_, remove_fuzz=True, force_severity=True,
                                     debug=self.debug, **kwargs)
            elif isinstance(answer.object, Distortion):
                pass
            elif isinstance(answer.object, (Aggregate, Portfolio)) and update is False:
                pass
            else:
                logger.warning(f'Unexpected: output kind is {type(answer.object)}. (expr/number?)')
                pass

        if len(rv) == 1:
            # only one output...just return that
            # retun object if it exists, otherwise the ans namedtuple
            for answer in rv:
                if answer.object is None:
                    return answer
                else:
                    return answer.object
        else:
            # multiple outputs, see if there is just one portfolio...this is not ideal?!
            ports_found = 0
            for answer in rv:
                if answer.kind == 'port':
                    ports_found += 1
            if ports_found == 1:
                # if only one, it must be answer
                if answer.object is None:
                    return answer
                else:
                    return answer.object
        # in all other cases, return the full list
        return rv

    __call__ = build

    def interpreter_file(self, *, filename='', where=''):
        """
        Run a suite of test programs. For detailed analysis, run_one.
        filename is a string or Path. If a csv it is read into
        a dataframe, with the first column used as index. If it
        is an agg file (e.g. an agg database), it is preprocessed
        to remove comments and replace \\n\\t agg with a space, then
        split on new lines and converted to a dataframe.
        Other file formats are rejected.

        These methods are called interpreter\_... rather than
        interpret\_... because they are for testing and debugging
        the interpreter, not for actually interpreting anything!

        """
        if filename == '':
            filename = Path.home() / 'aggregate/tests/test_suite.csv'
        elif type(filename) == str:
            filename = Path(filename)
        if filename.suffix == '.csv':
            df = pd.read_csv(filename, index_col=0)
        elif filename.suffix == '.agg':
            txt = filename.read_text(encoding='utf-8')
            stxt = re.sub('\n\tagg', ' agg', txt, flags=re.MULTILINE)
            stxt = stxt.split('\n')
            stxt = [i for i in stxt if len(i) and i[0] != '#']
            df = pd.DataFrame(stxt, columns=['program'])
        else:
            raise ValueError(f'File suffix must be .csv or .agg, not {filename.suffix}')
        if where != '':
            df = df.loc[df.index.str.match(where)]
        # add One severity if not input
        # if txt.find('sev One dsev [1]') < 0:
        #     logger.info('Adding One to knowledge.')
        self.write('sev One dsev [1]')
        return self._interpreter_work(df.iterrows())

    def interpreter_line(self, program, name='one off', debug=True):
        """
        Interpret single line of code  in debug mode.
        name is index of output
        """

        return self._interpreter_work(iter([(name, program)]), debug=debug)

    def interpreter_list(self, program_list):
        """
        Interpret elements in a list in debug mode.
        """
        return self._interpreter_work(list(enumerate(program_list)), debug=True)

    def _interpreter_work(self, iterable, debug=False):
        """
        Do all the work for the test, allows input to be marshalled into the tester
        in different ways. Unlike production interpret_program, runs one line at a time.
        Each line is preprocessed and then run through a clean parser, and the output
        analyzed.

        Last column, program as input is only changed if the preprocessor changes the program

        :return: DataFrame
        """
        lexer = UnderwritingLexer()
        parser = UnderwritingParser(self.safe_lookup, debug)
        ans = {}
        errs = 0
        no_errs = 0
        # detect non-trivial change
        f = lambda x, y: 'same' if x.replace(' ', '') == y.replace(' ', '').replace('\t', '') else y
        for test_name, program in iterable:
            if type(program) != str:
                program_in = program[0]
                program = lexer.preprocess(program_in)
            else:
                program_in = program
                program = lexer.preprocess(program_in)
            err = 0
            if len(program) == 1:
                program = program[0]
                try:
                    # print(program)
                    kind, name, spec = parser.parse(lexer.tokenize(program))
                except (ValueError, TypeError) as e:
                    errs += 1
                    err = 1
                    kind = program.split()[0]
                    # get something to say about the error
                    ea = getattr(e, 'args', None)
                    if ea is not None:
                        # t = getattr(ea[0], 'type', ea[0])
                        # v = getattr(ea[0], 'value', ea[0])
                        i = getattr(ea[0], 'index', 0)
                        if type(i) != int:  i = 0
                        # print(i, ea)
                        txt = program[0:i] + f'>>>' + program[i:]
                        name = 'parse error'
                    else:
                        txt = str(e)
                        name = 'other error'
                    spec = txt
                else:
                    no_errs += 1
                ans[test_name] = [kind, err, name, spec, program, f(program, program_in)]
            elif len(program) > 1:
                logger.info(f'{program_in} preprocesses to {len(program)} lines; not processing.')
                logger.info(program)
                ans[test_name] = ['multiline', err, None, None, program, program_in]
            else:
                logger.info(f'{program_in} preprocesses to a blank line; ignoring.')
                ans[test_name] = ['blank', err, None, None, program, program_in]

        df_out = pd.DataFrame(ans,
                              index=['kind', 'error', 'name', 'output', 'preprocessed program',
                                     'program']).T
        df_out.index.name = 'index'
        return df_out

    def qshow(self, regex):
        """
        Wrapper for show to just list elements in knowledge that match ``regex``.

        Either returns a dataframe or prints the dataframe.

        """
        return self.show(regex, kind='', plot=False, describe=False, return_df=True)

    def qlist(self, regex):
        qd(self.show(regex, kind='', plot=False, describe=False),
           line_width=160, max_colwidth=60, justify='left')

    def show(self, regex, kind='', plot=True, describe=True, logger_level=30, return_df=False):
        """
        Create from knowledge by name or match to name.
        Optionally plot. Returns the created object plus dataframe with more detailed information.
        ??How diff from describe??
        Allows exploration of pre-loaded databases.

        Eg ``regex = "A.*[234]`` for A...2, 3 and 4.

        See ``qshow`` for a wrapper that just returns the matches, with no object
        creation or plotting.

        Examples.
        ::

            from aggregate.utilities import pprint
            # pretty print all prgrams starting A; no object creation
            build.show('^A.*', 'agg', False, False).program.apply(pprint);

            # build and plot A..234
            ans, df = build.show('^A.*')

        :param regex: for filtering name
        :param kind: the kind of object, port, agg, etc.
        :param plot:    if True, plot   (default True)
        :param describe: if True, print the describe dataframe
        :param logger_level: work silently!
        :return: dictionary of created objects and DataFrame with info about each.
        """
        # too painful getting the one thing out!
        ans = []

        # temp logger level
        lm = LoggerManager(logger_level)

        if kind is None or kind == '':
            df = self.knowledge.droplevel('kind').filter(regex=regex, axis=0).copy()
        else:
            df = self.knowledge.loc[kind].filter(regex=regex, axis=0).copy()

        if plot is False and describe is False:
            # just act like a filtered listing on knowledge
            return df.sort_values('name')

        # added detail columns
        df['log2'] = 0
        df['bs'] = 0.
        df['agg_m'] = 0.
        df['agg_cv'] = 0.
        df['agg_sd'] = 0.
        df['emp_m'] = 0.
        df['emp_cv'] = 0.
        df['emp_sd'] = 0.

        for n, row in df.iterrows():
            p = row.program
            try:
                a = self.build(p)
                ans.append(a)
            except NotImplementedError:
                logger.error(f'skipping {n}...element not implemented')
            else:
                if describe:
                    # print('DecL Program:\n')
                    a.pprogram
                    # print('\n')
                    qd(a)
                if plot is True:
                    a.plot(figsize=(8, 2.4))
                    # print('\nDensity and Quantiles')
                    print()
                    show_fig(a.figure, format='svg')
                if describe:
                    print('\n')
                # info
                if isinstance(a, Portfolio):
                    m, cv = a.describe.loc[('total', 'Agg'), ['Est E[X]', 'Est CV(X)']]
                elif isinstance(a, Aggregate):
                    m, cv = a.describe.loc['Agg', ['Est E[X]', 'Est CV(X)']]
                else:
                    m = cv = np.nan
                df.loc[n, ['log2', 'bs', 'agg_m', 'agg_cv', 'agg_sd',
                            'emp_m', 'emp_cv', 'emp_sd']] = (a.log2, a.bs, a.agg_m, a.agg_cv,
                                                             a.agg_sd, m, cv, '')
        # if only one item, return it...much easier to use
        if len(ans) == 1: ans = a
        if return_df:
            return ans, df

    def dir(self, filter=''):
        """
        List all agg databases in site and default directories.
        If entries is True then read them and return named objects.

        :param filter:  glob filter for filename; .agg is added

        """

        if filter=='':
            filter = '*.agg'
        else:
            filter += '.agg'

        entries = []

        for dn, d in zip(['site', 'default'], [self.site_dir, self.default_dir]):
            for fn in d.glob(filter):
                txt = fn.read_text(encoding='utf-8')
                stxt = txt.split('\n')
                for r in stxt:
                    rs = r.split(' ')
                    if rs[0] in ['agg', 'port', 'dist', 'distortion', 'sev']:
                        entries.append([dn, fn.name] + rs[:2])

        ans = pd.DataFrame(entries, columns=['Directory', 'Database', 'kind', 'name'])
        return ans


# exported instance
# self = dbuild = None
logger_level(30)
build = Underwriter(databases='test_suite', update=True, debug=False, log2=16)
debug_build = Underwriter(name='Debug', update=True, debug=True, log2=13)
